import unittest
import numpy as np
from pyspedas.cotrans_tools.quaternions import (
    qslerp,
    qcompose,
    qconj,
    mtoq,
    qtom,
    qvalidate,
    qmult,
    qdecompose,
)


class Qtests(unittest.TestCase):
    def test_qslerp(self):
        m1 = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        m2 = np.array([[-1, 0, 0], [0, -1, 0], [0, 0, 1]])
        qin = np.transpose(np.array([mtoq(m1), mtoq(m2)])).reshape([4, 2]).transpose()
        x1 = np.array([0.0, 1.0])
        x2 = np.array([0.0, 1.0 / 6.0, 1.0 / 3.0, 1.0 / 2.0, 2.0 / 3.0, 5.0 / 6.0, 1.0])
        qout = qslerp(qin, x1, x2)
        qout_geo = qslerp(qin, x1, x2, geometric=True)
        mout = qtom(qout)
        qcomp = qcompose(np.array([[1, 2, 3]]), np.array([4]), free=False)
        self.assertTrue(len(mout) > 0)
        self.assertTrue(len(qcomp) > 0)
        self.assertTrue(np.abs(np.sum(qout - qout_geo)) < 1e-6)

    def test_errors(self):
        qs1 = np.ones((7, 5))
        qs2 = np.ones((8, 4))
        qs3 = np.ones(4)
        self.assertTrue(qvalidate(qs1, "qs1", "qslerp") == -1)
        self.assertTrue(qmult(qs1, qs2) == -1)
        self.assertTrue(qdecompose(np.array(1)) == -1)
        self.assertTrue(qdecompose(1) == -1)
        self.assertTrue(qconj(1) == -1)
        self.assertEqual(qslerp(qs1, qs2, qs3), -1)


if __name__ == "__main__":
    unittest.main()



"""
Compute power spectra for data.

Notes
-----
Similar to dpwrspc.pro in IDL SPEDAS.

"""

import logging
import numpy as np


def dpwrspc(
    time,
    quantity,
    nboxpoints=256,
    nshiftpoints=128,
    bin=3,
    tbegin=-1.0,
    tend=-1.0,
    noline=False,
    nohanning=False,
    notperhz=False,
    notmvariance=False,
    tm_sensitivity=None,
):
    """
    Compute power spectra.

    Parameters
    ----------
    time: list of float
        Time array.
    quantity: list of float
        Data array.
    nboxpoints: int, optional
        The number of points to use for the hanning window.
        The default is 256.
    nshiftpoints: int, optional
        The number of points to shift for each spectrum.
        The default is 128.
    bin: int, optional
        Size for binning of the data along the frequency domain.
        The default is 3.
    tbegin: float, optional
        Start time for the calculation.
        If -1.0, the start time is the first time in the time array.
    tend: float, optional
        End time for the calculation.
        If -1.0, the end time is the last time in the time array.
    noline: bool, optional
        If True, no straight line is subtracted.
        The default is False.
    nohanning: bool, optional
        If True, no hanning window is applied to the input.
        The default is False.
    notperhz: bool, optional
        If True, the output units are the square of the input units.
        The default is False.
    notmvariance: bool, optional
        If True, replace output spectrum for any windows that have variable
        cadence with NaNs.
        The default is False.
    tm_sensitivity: float, optional
        If noTmVariance is set, this number controls how much of a dt anomaly
        is accepted.
        The default is None.

    Returns
    -------
    tuple

        tdps: array of float
            The time array for the dynamic power spectrum, the center time of the
            interval used for the spectrum.
        fdps: array of float
            The frequency array (units =1/time units).
        dps: array of float
            The power spectrum, (units of quantity)^2/frequency_units.

    Examples
    --------

        >>> # Compute the power spectrum of a given time series
        >>> import numpy as np
        >>> time = range(3000)
        >>> quantity = np.random.random(3000)
        >>> power = pyspedas.dpwrspc(time, quantity)
    """
    tdps, fdps, dps = np.array(-1.0), np.array(-1.0), np.array(-1.0)
    tdps0, fdps0, dps0 = np.array(-1.0), np.array(-1.0), np.array(-1.0)

    if nohanning is False:
        window = np.array(np.hanning(nboxpoints), dtype=np.float64)
    else:
        window = np.ones(nboxpoints)

    time = np.array(time, dtype=np.float64)
    quantity = np.array(quantity, dtype=np.float64)

    if tbegin == -1.0:
        tbegin = time[0]
    if tend == -1.0:
        tend = time[-1]

    # Find the time array that correspond to the start and end times
    tbegin_idx = np.where(time >= tbegin)[0][0]
    tend_idx = np.where(time <= tend)[0][-1]
    if tend_idx - tbegin_idx < nboxpoints:
        logging.error("Not enough points for a calculation")
        return tdps0, fdps0, dps0
    time = time[tbegin_idx : tend_idx + 1]
    quantity = quantity[tbegin_idx : tend_idx + 1]

    # remove NaNs from the data
    where_finite = np.where(np.isnan(quantity) == False)
    quantity2process = quantity[where_finite[0]]
    times2process = time[where_finite[0]]
    nboxpnts = int(nboxpoints)
    nshiftpnts = nshiftpoints

    totalpoints = len(times2process)
    nspectra = int((totalpoints - nboxpnts / 2.0) / nshiftpnts)

    # test nspectra, if the value of nshiftpnts is much smaller than
    # nboxpnts/2 strange things happen

    nbegin = np.array([nshiftpnts * i for i in range(nspectra)], dtype=np.int64)
    nend = nbegin + nboxpnts

    okspec = np.where(nend <= totalpoints - 1)

    if len(okspec[0]) <= 0:
        logging.error("Not enough points for a calculation")
        return tdps0, fdps0, dps0
    else:
        nspectra = len(okspec[0])

    tdps = np.zeros(nspectra)
    nfreqs = int(int(nboxpnts / 2) / bin)

    if nfreqs <= 1:
        logging.error("Not enough frequencies for a calculation")
        return tdps0, fdps0, dps0

    dps = np.zeros([nspectra, nfreqs])
    fdps = np.zeros([nspectra, nfreqs])

    # Main calculation loop
    for nthspectrum in range(0, nspectra):
        nbegin = int(nthspectrum * nshiftpnts)
        nend = nbegin + nboxpnts

        if nend <= totalpoints:
            t = times2process[nbegin:nend]
            t0 = t[0]
            t = t - t0
            x = quantity2process[nbegin:nend]

            # Use center time
            tdps[nthspectrum] = (times2process[nbegin] + times2process[nend - 1]) / 2.0

            if noline is False:
                coef = np.polyfit(t, x, 1)
                poly1d_fn = np.poly1d(coef)
                line = poly1d_fn(t)
                x = x - line

            if nohanning is False:
                x = x * window

            bign = nboxpnts

            if bign % 2 != 0:
                logging.warning(
                    "dpwrspc: needs an even number of data points, dropping last point..."
                )
                t = t[0 : bign - 1]
                x = x[0 : bign - 1]
                bign = bign - 1

            n_tm = len(t)

            # time variance can break power spectrum
            # this keyword skips over those gaps
            if notmvariance and n_tm > 1:
                if tm_sensitivity is not None:
                    tmsn = tm_sensitivity
                else:
                    tmsn = 100.0

                tdiff = t[1:n_tm] - t[0 : n_tm - 1]
                med_diff = np.median(tdiff)

                idx = np.where(np.abs(tdiff / med_diff - 1) > 1.0 / tmsn)

                if len(idx[0]) > 0:
                    dps[nthspectrum, :] = float("nan")
                    fdps[nthspectrum, :] = float("nan")

                    continue

            # following Numerical recipes in Fortran, p. 421, sort of...
            # (actually following the IDL implementation)
            k = np.array(range(int(bign / 2) + 1))
            tres = np.median(t[1 : len(t)] - t[0 : len(t) - 1])
            fk = k / (bign * tres)

            xs2 = np.abs(np.fft.fft(x)) ** 2

            pwr = np.zeros(int(bign / 2 + 1))
            pwr[0] = xs2[0] / bign**2
            ptmp = 1 + np.array(range(int(bign / 2 - 1)))
            pwr[1 : int(bign / 2)] = (
                xs2[1 : int(bign / 2)] + xs2[bign - ptmp]
            ) / bign**2
            pwr[int(bign / 2)] = xs2[int(bign / 2)] / bign**2

            if nohanning is False:
                wss = float(bign) * float(np.sum(window**2))
                pwr = bign**2 * pwr / wss

            dfreq = bin * (fk[1] - fk[0])

            npwr = len(pwr) - 1
            nfinal = int(npwr / bin)
            iarray = np.array(range(nfinal))
            power = np.zeros(nfinal)

            # Note: zeroth point includes zero freq. power.
            freqcenter = (fk[iarray * bin + 1] + fk[iarray * bin + bin]) / 2.0

            for i in range(bin):
                power = power + pwr[iarray * bin + i + 1]

            if notperhz is False:
                power = power / dfreq

            dps[nthspectrum, :] = power
            fdps[nthspectrum, :] = freqcenter

    return tdps, fdps, dps



import logging
import os

from .config import CONFIG
from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def month_to_num(month_string):
    """
    Converts a three-letter month abbreviation to its corresponding two-digit number.

    Parameters
    ----------
    month_string : str
        The three-letter month abbreviation (e.g., "JAN", "FEB", etc.).

    Returns
    -------
    str
        The two-digit number corresponding to the month abbreviation.

    Raises
    ------
    ValueError
        If the month string is not a valid abbreviation.

    Examples
    --------
    >>> month_to_num("JAN")
    '01'
    >>> month_to_num("FEB")
    '02'
    """

    month_to_num_dict = {
        "JAN": "01",
        "FEB": "02",
        "MAR": "03",
        "APR": "04",
        "MAY": "05",
        "JUN": "06",
        "JUL": "07",
        "AUG": "08",
        "SEP": "09",
        "OCT": "10",
        "NOV": "11",
        "DEC": "12",
    }

    if month_string in month_to_num_dict.keys():
        return month_to_num_dict[month_string]
    else:
        raise ValueError("Month string is not valid")


def orbit_time(begin_orbit, end_orbit=None):
    """
    Calculate the corresponding time range for a given MAVEN orbit number or range of orbit numbers.

    Parameters
    ----------
    begin_orbit : int
        The beginning orbit number.
    end_orbit : int, optional
        The ending orbit number. If not provided, the time range will be calculated for a single orbit.

    Returns
    -------
    list
        A list containing the beginning and ending time in the format [begin_time, end_time].
        If the orbit numbers are not found, [None, None] will be returned.
    """

    toolkit_path = CONFIG["local_data_dir"]
    sep = "/" if is_fsspec_uri(toolkit_path) else os.path.sep
    orbit_files_path = sep.join([toolkit_path, "orbitfiles"])
    orb_file = sep.join([orbit_files_path, "merged_maven_orbits.orb"])
    logging.info("Getting orbit info from file %s", orb_file)
    if is_fsspec_uri(toolkit_path):
        protocol, path = toolkit_path.split("://")
        fs = fsspec.filesystem(protocol)

        fileobj = fs.open(orb_file, "r")
    else:
        fileobj = open(orb_file, "r")

    with fileobj as f:
        if end_orbit is None:
            end_orbit = begin_orbit
        orbit_num = []
        time = []
        f.readline()
        f.readline()
        for line in f:
            line = line[0:28]
            line = line.split(" ")
            line = [x for x in line if x != ""]
            orbit_num.append(int(line[0]))
            time.append(
                line[1] + "-" + month_to_num(line[2]) + "-" + line[3] + "T" + line[4]
            )
        try:
            if orbit_num.index(begin_orbit) > len(time) or orbit_num.index(
                end_orbit
            ) + 1 > len(time):
                logging.warning(
                    "Orbit numbers not found.  Please choose a number between 1 and %s.",
                    orbit_num[-1],
                )
                return [None, None]
            else:
                begin_time = time[orbit_num.index(begin_orbit)]
                end_time = time[orbit_num.index(end_orbit) + 1]
        except ValueError:
            logging.warning('Problem getting begin and end times for orbits %d and %d', begin_orbit, end_orbit+1)
            return [None, None]
    return [begin_time, end_time]



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(
    trange=["2018-11-5", "2018-11-6"],
    probe=["noaa19"],
    instrument="sem",
    datatype="*",
    prefix="",
    suffix="",
    get_support_data=False,
    ncei_server=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    Load POES Space Environment Monitor data.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD', 'YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss', 'YYYY-MM-DD/hh:mm:ss'].
        Default is ['2018-11-5', '2018-11-6'].
    probe : str or list of str, optional
        POES spacecraft name(s); e.g., metop1, metop2, noaa15, noaa16,
        noaa18, noaa19. Default is ['noaa19'].
    instrument : str, optional
        Name of the instrument. Default is 'sem'.
    datatype : str, optional
        This variable is unused. It is reserved for future use. Default is '*'.
    prefix : str, optional
        The tplot variable names will be given this prefix. By default, no prefix is added.
        Default is ''.
    suffix : str, optional
        The tplot variable names will be given this suffix. By default, no suffix is added.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data"
        will be loaded into tplot. By default, only loads in data with a
        "VAR_TYPE" attribute of "data". Default is False.
    ncei_server : bool, optional
        Set this flag to download data from the NCEI NOAA server instead of the default SPDF server.
        The NCEI server contains older data which is not available on the SPDF server. Default is False.
    varformat : str, optional
        The file variable formats to load into tplot. Wildcard character
        "*" is accepted. By default, all variables are loaded in. Default is False.
    varnames : list of str, optional
        List of variable names to load (if not specified, all data variables are loaded).
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables. Default is False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables. Default is False.
    no_update : bool, optional
        If set, only load data from your local cache. Default is False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword. Default is False.
    force_download : bool, optional
        Download file even if local version is more recent than server version. Default is False.

    Returns
    -------
    dict or list
        List of tplot variables created.

    Examples
    --------
    >>> from pyspedas.projects.poes.load import load
    >>> vars = load(trange=['2013-11-5', '2013-11-6'])
    >>> print(vars)
    """

    if not isinstance(probe, list):
        probe = [probe]

    out_files = []
    tvars = []

    for prb in probe:
        if ncei_server:
            remote_path = CONFIG["ncei_remote_data_dir"]
            num = prb[-2:]
            pathformat = "%Y/" + prb + "/poes_n" + num + "_%Y%m%d.cdf"
        else:
            remote_path = CONFIG["remote_data_dir"]
            if instrument == "sem":
                pathformat = (
                    prb
                    + "/sem2_fluxes-2sec/%Y/"
                    + prb
                    + "_poes-sem2_fluxes-2sec_%Y%m%d_v??.cdf"
                )

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)

        files = download(
            remote_file=remote_names,
            remote_path=remote_path,
            local_path=CONFIG["local_data_dir"],
            no_download=no_update,
            force_download=force_download,
        )
        if files is not None:
            for file in files:
                out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(
        out_files,
        prefix=prefix,
        suffix=suffix,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        notplot=notplot,
    )

    if notplot:
        return tvars

    if time_clip:
        tclip(tvars, trange[0], trange[1], suffix="", overwrite=True)

    return tvars



from .load import load

# This routine was originally in st5/__init__.py, until being moved to its own file.
# Please refer to __init__.py if you need to see the revision history before it was moved.

def mag_postprocessing(variables):
    """
    Placeholder for MAG post-processing
    """
    return variables



import logging
import calendar
import numpy as np
from .utilities import maven_kp_l2_regex
from .utilities import maven_param_list
from .utilities import remove_inst_tag
from .utilities import (
    get_latest_files_from_date_range,
    get_latest_iuvs_files_from_date_range,
)
from .utilities import get_header_info
from .orbit_time import orbit_time
from .read_iuvs_file import read_iuvs_file
import pyspedas
from pyspedas.tplot_tools import store_data, link, join_vec
from collections import OrderedDict
import builtins
import os

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def maven_kp_to_tplot(
    filename=None,
    input_time=None,
    instruments=None,
    insitu_only=False,
    specified_files_only=False,
    ancillary_only=False,
    notplot=False
):
    """
    Convert MAVEN insitu (KP) data files to tplot variables.

    Parameters
    ----------
    filename : str or list, optional
        Name of the file(s) to read in. Defaults to None.
    input_time : str or list or int, optional
        Time frame in which to search for downloaded files.
        Can be a single date string, a list of two date strings representing a time range, or an orbit number.
        Defaults to None.
    instruments : str or list, optional
        List of instruments to include. Defaults to None.
    insitu_only : bool, optional
        Flag to include only insitu data files. Defaults to False.
    specified_files_only : bool, optional
        If True, specifies that only filenames given in 'filename' should be read in. Defaults to False.
    ancillary_only : bool, optional
        If True, only the spacecraft and APP info will be loaded. Defaults to False.
    notplot: bool, optional
        If true, return the kp_insitu data as a dict rather than tplot variables. Defaults to False.

    Returns
    -------
    dict
        A dictionary containing up to all of the columns included in a MAVEN in-situ Key parameter data file.
    """

    import pandas as pd
    import re
    from datetime import datetime, timedelta
    from dateutil.parser import parse

    filenames = []
    iuvs_filenames = []
    param_dict = maven_param_list()

    if instruments is not None:
        if not isinstance(instruments, builtins.list):
            instruments = [instruments]

    if filename is None and input_time is None:
        logging.warning(
            "You must specify either a set of filenames to read in, or a time frame in which "
            "you want to search for downloaded files."
        )

    if ancillary_only:
        instruments = ["SPACECRAFT"]

    if filename is not None:
        if not isinstance(filename, builtins.list):
            filename = [filename]

        dates = []
        for file in filename:
            date = re.findall(r"_(\d{8})", file)[0]
            dates.append(date)
            if "iuvs" in file:
                iuvs_filenames.append(file)
            else:
                filenames.append(file)
        dates.sort()

        # To keep the rest of the code consistent, if someone gave a files, or files, to load, but no input_time,
        # go ahead and create an 'input_time'
        if input_time is None:
            if len(dates) == 1:
                input_time = (
                    str(dates[0][:4])
                    + "-"
                    + str(dates[0][4:6])
                    + "-"
                    + str(dates[0][6:])
                )

            else:
                beg_date = min(dates)
                end_date = max(dates)
                input_time = [
                    str(beg_date[:4])
                    + "-"
                    + str(beg_date[4:6])
                    + "-"
                    + str(beg_date[6:]),
                    str(end_date[:4])
                    + "-"
                    + str(end_date[4:6])
                    + "-"
                    + str(end_date[6:]),
                ]

    # Check for orbit num rather than time string
    if isinstance(input_time, builtins.list):
        if isinstance(input_time[0], int):
            input_time = orbit_time(input_time[0], input_time[1])
    elif isinstance(input_time, int):
        input_time = orbit_time(input_time)

    # Turn string input into datetime objects
    if isinstance(input_time, list):
        if len(input_time[0]) <= 10:
            input_time[0] = input_time[0] + " 00:00:00"
        if len(input_time[1]) <= 10:
            input_time[1] = input_time[1] + " 23:59:59"
        date1 = parse(input_time[0])
        date2 = parse(input_time[1])
    else:
        if len(input_time) <= 10:
            input_time += " 00:00:00"
        date1 = parse(input_time)
        date2 = date1 + timedelta(days=1)

    date1_unix = calendar.timegm(date1.timetuple())
    date2_unix = calendar.timegm(date2.timetuple())

    # Grab insitu and iuvs files for the specified/created date ranges
    date_range_filenames = get_latest_files_from_date_range(date1, date2)
    date_range_iuvs_filenames = get_latest_iuvs_files_from_date_range(date1, date2)

    # Add date range files to respective file lists if desired
    if not specified_files_only:
        filenames.extend(date_range_filenames)
        iuvs_filenames.extend(date_range_iuvs_filenames)

    if not date_range_filenames and not date_range_iuvs_filenames:
        if not filenames and not iuvs_filenames:
            logging.warning(
                "No files found for the input date range, and no specific filenames were given. Exiting."
            )
            return

    # Going to look for files between time frames, but as we might have already specified
    # certain files to load in, we don't want to load them in 2x... so doing a check for that here
    filenames = list(set(filenames))
    iuvs_filenames = list(set(iuvs_filenames))

    kp_regex, l2_regex = maven_kp_l2_regex()
    kp_insitu = []
    if filenames:
        # Get column names
        names, inst = [], []
        crus_name, crus_inst = [], []
        c_found = False
        r_found = False
        for f in filenames:
            if is_fsspec_uri(f):
                protocol, path = f.split("://")
                fs = fsspec.filesystem(protocol)

                basename = f.rstrip("/").split("/")[-1]
            else:
                basename = os.path.basename(f)
            if (
                kp_regex.match(basename).group("description") == "_crustal"
                and not c_found
            ):
                name, inss = get_header_info(f)
                # Strip off the first name for now (Time), and use that as the dataframe index.
                # Seems to make sense for now, but will it always?
                crus_name.extend(name[1:])
                crus_inst.extend(inss[1:])
                c_found = True
            elif (
                kp_regex.match(basename).group("description") == ""
                and not r_found
            ):
                name, ins = get_header_info(f)
                # Strip off the first name for now (Time), and use that as the dataframe index.
                # Seems to make sense for now, but will it always?
                names.extend(name[1:])
                inst.extend(ins[1:])
                r_found = True
        all_names = names + crus_name
        all_inst = inst + crus_inst

        # Break up dictionary into instrument groups
        (
            lpw_group,
            euv_group,
            swe_group,
            swi_group,
            sta_group,
            sep_group,
            mag_group,
            ngi_group,
            app_group,
            sc_group,
            crus_group,
        ) = ([], [], [], [], [], [], [], [], [], [], [])

        for i, j in zip(all_inst, all_names):
            if re.match("^LPW$", i.strip()):
                lpw_group.append(j)
            elif re.match("^LPW-EUV$", i.strip()):
                euv_group.append(j)
            elif re.match("^SWEA$", i.strip()):
                swe_group.append(j)
            elif re.match("^SWIA$", i.strip()):
                swi_group.append(j)
            elif re.match("^STATIC$", i.strip()):
                sta_group.append(j)
            elif re.match("^SEP$", i.strip()):
                sep_group.append(j)
            elif re.match("^MAG$", i.strip()):
                mag_group.append(j)
            elif re.match("^NGIMS$", i.strip()):
                ngi_group.append(j)
            elif re.match("^MODELED_MAG$", i.strip()):
                crus_group.append(j)
            elif re.match("^SPICE$", i.strip()):
                # NB Need to split into APP and SPACECRAFT
                if re.match("(.+)APP(.+)", j):
                    app_group.append(j)
                else:  # Everything not APP is SC in SPICE
                    # But do not include Orbit Num, or IO Flag
                    # Could probably stand to clean this line up a bit
                    if not re.match("(.+)(Orbit Number|Inbound Outbound Flag)", j):
                        sc_group.append(j)
            else:
                pass

        delete_groups = []
        if instruments is not None:
            if "LPW" not in instruments and "lpw" not in instruments:
                delete_groups += lpw_group
            if "MAG" not in instruments and "mag" not in instruments:
                delete_groups += mag_group
            if "EUV" not in instruments and "euv" not in instruments:
                delete_groups += euv_group
            if "SWI" not in instruments and "swi" not in instruments:
                delete_groups += swi_group
            if "SWE" not in instruments and "swe" not in instruments:
                delete_groups += swe_group
            if "NGI" not in instruments and "ngi" not in instruments:
                delete_groups += ngi_group
            if "SEP" not in instruments and "sep" not in instruments:
                delete_groups += sep_group
            if "STA" not in instruments and "sta" not in instruments:
                delete_groups += sta_group
            if "MODELED_MAG" not in instruments and "modeled_mag" not in instruments:
                delete_groups += crus_group

        # Read in all relavent data into a pandas dataframe called "temp"
        temp_data = []
        filenames.sort()
        for filename in filenames:
            # Determine number of header lines
            nheader = 0
            if is_fsspec_uri(filename):
                protocol, path = filename.split("://")
                fs = fsspec.filesystem(protocol)
                fo = fs.open(filename, "rt")
            else:
                fo = open(filename)
            with fo as f:
                for line in f:
                    if line.startswith("#"):
                        nheader += 1
                if is_fsspec_uri(filename):
                    protocol, path = filename.split("://")
                    fs = fsspec.filesystem(protocol)

                    basename = filename.rstrip("/").split("/")[-1]
                else:
                    basename = os.path.basename(filename)
                if (
                    kp_regex.match(basename).group("description")
                    == "_crustal"
                ):
                    temp_data.append(
                        pd.read_fwf(
                            filename,
                            skiprows=nheader,
                            index_col=0,
                            widths=[19] + len(crus_name) * [16],
                            names=crus_name,
                        )
                    )
                else:
                    temp_data.append(
                        pd.read_fwf(
                            filename,
                            skiprows=nheader,
                            index_col=0,
                            widths=[19] + len(names) * [16],
                            names=names,
                        )
                    )
                for i in delete_groups:
                    del temp_data[-1][i]

        temp_unconverted = pd.concat(temp_data, axis=0, sort=True)

        # Need to convert columns
        # This is kind of a hack, but I can't figure out a better way for now

        if (
            "SWEA.Electron Spectrum Shape" in temp_unconverted
            and "NGIMS.Density NO" in temp_unconverted
        ):
            temp = temp_unconverted.astype(
                dtype={
                    "SWEA.Electron Spectrum Shape": np.float64,
                    "NGIMS.Density NO": np.float64,
                }
            )
        elif (
            "SWEA.Electron Spectrum Shape" in temp_unconverted
            and "NGIMS.Density NO" not in temp_unconverted
        ):
            temp = temp_unconverted.astype(
                dtype={"SWEA.Electron Spectrum Shape": np.float64}
            )
        elif (
            "SWEA.Electron Spectrum Shape" not in temp_unconverted
            and "NGIMS.Density NO" in temp_unconverted
        ):
            temp = temp_unconverted.astype(dtype={"NGIMS.Density NO": np.float64})
        else:
            temp = temp_unconverted

        # Cut out the times not included in the date range
        time_unix = [
            calendar.timegm(datetime.strptime(i, "%Y-%m-%dT%H:%M:%S").timetuple())
            for i in temp.index
        ]
        start_index = 0
        for t in time_unix:
            if t >= date1_unix:
                break
            start_index += 1
        end_index = 0
        for t in time_unix:
            if t >= date2_unix:
                break
            end_index += 1

        # Assign the first-level only tags
        time_unix = time_unix[start_index:end_index]
        temp = temp[start_index:end_index]
        time = temp.index
        time_unix = pd.Series(time_unix)  # convert into Series for consistency
        time_unix.index = temp.index

        if "SPICE.Orbit Number" in list(temp):
            orbit = temp["SPICE.Orbit Number"]
        else:
            orbit = None
        if "SPICE.Inbound Outbound Flag" in list(temp):
            io_flag = temp["SPICE.Inbound Outbound Flag"]
        else:
            io_flag = None

        # Build the sub-level DataFrames for the larger dictionary/structure
        app = temp[app_group]
        spacecraft = temp[sc_group]
        if instruments is not None:
            if "LPW" in instruments or "lpw" in instruments:
                lpw = temp[lpw_group]
            else:
                lpw = None
            if "MAG" in instruments or "mag" in instruments:
                mag = temp[mag_group]
            else:
                mag = None
            if "EUV" in instruments or "euv" in instruments:
                euv = temp[euv_group]
            else:
                euv = None
            if "SWE" in instruments or "swe" in instruments:
                swea = temp[swe_group]
            else:
                swea = None
            if "SWI" in instruments or "swi" in instruments:
                swia = temp[swi_group]
            else:
                swia = None
            if "NGI" in instruments or "ngi" in instruments:
                ngims = temp[ngi_group]
            else:
                ngims = None
            if "SEP" in instruments or "sep" in instruments:
                sep = temp[sep_group]
            else:
                sep = None
            if "STA" in instruments or "sta" in instruments:
                static = temp[sta_group]
            else:
                static = None
            if "MODELED_MAG" in instruments or "modeled_mag" in instruments:
                crus = temp[crus_group]
            else:
                crus = None
        else:
            lpw = temp[lpw_group]
            euv = temp[euv_group]
            swea = temp[swe_group]
            swia = temp[swi_group]
            static = temp[sta_group]
            sep = temp[sep_group]
            mag = temp[mag_group]
            ngims = temp[ngi_group]
            crus = temp[crus_group]

        # Strip out the duplicated instrument part of the column names
        # (this is a bit hardwired and can be improved)
        for i in [lpw, euv, swea, swia, sep, static, ngims, mag, crus, app, spacecraft]:
            if i is not None:
                i.columns = remove_inst_tag(i)

        if lpw is not None:
            lpw = lpw.rename(index=str, columns=param_dict)
        if euv is not None:
            euv = euv.rename(index=str, columns=param_dict)
        if swea is not None:
            swea = swea.rename(index=str, columns=param_dict)
        if swia is not None:
            swia = swia.rename(index=str, columns=param_dict)
        if sep is not None:
            sep = sep.rename(index=str, columns=param_dict)
        if static is not None:
            static = static.rename(index=str, columns=param_dict)
        if ngims is not None:
            ngims = ngims.rename(index=str, columns=param_dict)
        if mag is not None:
            mag = mag.rename(index=str, columns=param_dict)
        if crus is not None:
            crus = crus.rename(index=str, columns=param_dict)
        if app is not None:
            app = app.rename(index=str, columns=param_dict)
        if spacecraft is not None:
            spacecraft = spacecraft.rename(index=str, columns=param_dict)

        if orbit is not None and io_flag is not None:
            # Do not forget to save units
            # Define the list of first level tag names
            tag_names = [
                "TimeString",
                "Time",
                "Orbit",
                "IOflag",
                "LPW",
                "EUV",
                "SWEA",
                "SWIA",
                "STATIC",
                "SEP",
                "MAG",
                "NGIMS",
                "MODELED_MAG",
                "APP",
                "SPACECRAFT",
            ]

            # Define list of first level data structures
            data_tags = [
                time,
                time_unix,
                orbit,
                io_flag,
                lpw,
                euv,
                swea,
                swia,
                static,
                sep,
                mag,
                ngims,
                crus,
                app,
                spacecraft,
            ]
        else:
            # Do not forget to save units
            # Define the list of first level tag names
            tag_names = [
                "TimeString",
                "Time",
                "LPW",
                "EUV",
                "SWEA",
                "SWIA",
                "STATIC",
                "SEP",
                "MAG",
                "NGIMS",
                "MODELED_MAG",
                "APP",
                "SPACECRAFT",
            ]

            # Define list of first level data structures
            data_tags = [
                time,
                time_unix,
                lpw,
                euv,
                swea,
                swia,
                static,
                sep,
                mag,
                ngims,
                crus,
                app,
                spacecraft,
            ]

        kp_insitu = OrderedDict(zip(tag_names, data_tags))

    # Now for IUVS
    kp_iuvs = []
    if not insitu_only and iuvs_filenames:
        for file in iuvs_filenames:
            kp_iuvs.append(read_iuvs_file(file))

    if notplot:
        return kp_insitu

    if not kp_iuvs:
        return tplot_varcreate(kp_insitu)
    else:
        # return kp_insitu, kp_iuvs
        insitu_vars = tplot_varcreate(kp_insitu)
        # The kp_iuvs structure can't be converted using tplot_varcreate.  Not clear
        # what needs to be done in order to use them. We'll punt on it for now and just
        # return the insitu variables.
        # iuvs_vars = tplot_varcreate(kp_iuvs)
        return insitu_vars


def tplot_varcreate(insitu):
    """Creates tplot variables from the insitu variable"""
    # initialize each instrument
    created_vars = []
    for obs in insitu["SPACECRAFT"]:
        obs_specific = "mvn_kp::spacecraft::" + obs.lower()
        try:
            store_data(
                obs_specific, data={"x": insitu["Time"], "y": insitu["SPACECRAFT"][obs]}
            )
            created_vars.append(obs_specific)
        except:
            pass

    # Join together the matricies and remove the individual points
    join_vec(
        [
            "mvn_kp::spacecraft::t11",
            "mvn_kp::spacecraft::t12",
            "mvn_kp::spacecraft::t13",
            "mvn_kp::spacecraft::t21",
            "mvn_kp::spacecraft::t22",
            "mvn_kp::spacecraft::t23",
            "mvn_kp::spacecraft::t31",
            "mvn_kp::spacecraft::t32",
            "mvn_kp::spacecraft::t33",
        ],
        newname="mvn_kp::geo_to_mso_matrix",
    )

    pyspedas.tplot_tools.del_data(
        [
            "mvn_kp::spacecraft::t11",
            "mvn_kp::spacecraft::t12",
            "mvn_kp::spacecraft::t13",
            "mvn_kp::spacecraft::t21",
            "mvn_kp::spacecraft::t22",
            "mvn_kp::spacecraft::t23",
            "mvn_kp::spacecraft::t31",
            "mvn_kp::spacecraft::t32",
            "mvn_kp::spacecraft::t33",
        ]
    )

    join_vec(
        [
            "mvn_kp::spacecraft::spacecraft_t11",
            "mvn_kp::spacecraft::spacecraft_t12",
            "mvn_kp::spacecraft::spacecraft_t13",
            "mvn_kp::spacecraft::spacecraft_t21",
            "mvn_kp::spacecraft::spacecraft_t22",
            "mvn_kp::spacecraft::spacecraft_t23",
            "mvn_kp::spacecraft::spacecraft_t31",
            "mvn_kp::spacecraft::spacecraft_t32",
            "mvn_kp::spacecraft::spacecraft_t33",
        ],
        newname="mvn_kp::spacecraft_to_mso_matrix",
    )

    pyspedas.tplot_tools.del_data(
        [
            "mvn_kp::spacecraft::spacecraft_t11",
            "mvn_kp::spacecraft::spacecraft_t12",
            "mvn_kp::spacecraft::spacecraft_t13",
            "mvn_kp::spacecraft::spacecraft_t21",
            "mvn_kp::spacecraft::spacecraft_t22",
            "mvn_kp::spacecraft::spacecraft_t23",
            "mvn_kp::spacecraft::spacecraft_t31",
            "mvn_kp::spacecraft::spacecraft_t32",
            "mvn_kp::spacecraft::spacecraft_t33",
        ]
    )

    created_vars.remove("mvn_kp::spacecraft::t11")
    created_vars.remove("mvn_kp::spacecraft::t12")
    created_vars.remove("mvn_kp::spacecraft::t13")
    created_vars.remove("mvn_kp::spacecraft::t21")
    created_vars.remove("mvn_kp::spacecraft::t22")
    created_vars.remove("mvn_kp::spacecraft::t23")
    created_vars.remove("mvn_kp::spacecraft::t31")
    created_vars.remove("mvn_kp::spacecraft::t32")
    created_vars.remove("mvn_kp::spacecraft::t33")
    created_vars.remove("mvn_kp::spacecraft::spacecraft_t11")
    created_vars.remove("mvn_kp::spacecraft::spacecraft_t12")
    created_vars.remove("mvn_kp::spacecraft::spacecraft_t13")
    created_vars.remove("mvn_kp::spacecraft::spacecraft_t21")
    created_vars.remove("mvn_kp::spacecraft::spacecraft_t22")
    created_vars.remove("mvn_kp::spacecraft::spacecraft_t23")
    created_vars.remove("mvn_kp::spacecraft::spacecraft_t31")
    created_vars.remove("mvn_kp::spacecraft::spacecraft_t32")
    created_vars.remove("mvn_kp::spacecraft::spacecraft_t33")

    inst_list = ["EUV", "LPW", "STATIC", "SWEA", "SWIA", "MAG", "SEP", "NGIMS"]
    for instrument in inst_list:
        # for each observation for each instrument
        if instrument in insitu:
            if insitu[instrument] is not None:
                for obs in insitu[instrument]:
                    # create variable name
                    obs_specific = "mvn_kp::" + instrument.lower() + "::" + obs.lower()
                    try:
                        # store data in tplot variable
                        store_data(
                            obs_specific,
                            data={"x": insitu["Time"], "y": insitu[instrument][obs]},
                        )
                        created_vars.append(obs_specific)
                        link(
                            obs_specific,
                            "mvn_kp::spacecraft::altitude",
                            link_type="alt",
                        )
                        link(
                            obs_specific, "mvn_kp::spacecraft::mso_x", link_type="x"
                        )
                        link(
                            obs_specific, "mvn_kp::spacecraft::mso_y", link_type="y"
                        )
                        link(
                            obs_specific, "mvn_kp::spacecraft::mso_z", link_type="z"
                        )
                        link(
                            obs_specific, "mvn_kp::spacecraft::geo_x", link_type="geo_x"
                        )
                        link(
                            obs_specific, "mvn_kp::spacecraft::geo_y", link_type="geo_y"
                        )
                        link(
                            obs_specific, "mvn_kp::spacecraft::geo_z", link_type="geo_z"
                        )
                        link(
                            obs_specific,
                            "mvn_kp::spacecraft::sub_sc_longitude",
                            link_type="lon",
                        )
                        link(
                            obs_specific,
                            "mvn_kp::spacecraft::sub_sc_latitude",
                            link_type="lat",
                        )
                    except:
                        pass
    return created_vars



import numpy as np
from scipy import interpolate

def interpol(
    data,
    data_times,
    out_times,
    fill_value="extrapolate"):

    """
    Interpolate data.

    Parameters
    ----------
    data: list of float
        Data array.
    data_times: list of float
        Time array.
    out_times: list of float
        Time array to interpolate to
    fill_value: str, optional
        Default: "extrapolate"

    Returns
    -------
    Interpolated data

    Examples
    --------

        >>> import pyspedas
        >>> x1 = [0, 4, 8, 12]
        >>> time1 = [pyspedas.time_float("2020-01-01") + i for i in x1]
        >>> x2=[0, 2, 4, 6, 8, 10, 12]
        >>> y1=[10, 20, 30, 40]
        >>> time2 = [pyspedas.time_float("2020-01-01") + i for i in x2]
        >>> pyspedas.interpol(y1, time1, time2)
    """
    if isinstance(data, list):
        data = np.array(data)

    if len(data.shape) == 2:
        out = np.empty((len(out_times), len(data[0, :])))
        for data_idx in np.arange(len(data[0, :])):
            interpfunc = interpolate.interp1d(data_times, data[:, data_idx], kind='linear', bounds_error=False, fill_value=fill_value)
            out[:, data_idx] = interpfunc(out_times)
        return out
    else:
        interpfunc = interpolate.interp1d(data_times, data, kind='linear', bounds_error=False, fill_value=fill_value)
        return interpfunc(out_times)



import unittest
import pyspedas
from pyspedas.tplot_tools import data_exists, del_data, get_data
import numpy as np


class LoadTestCases(unittest.TestCase):
    def test_load_kp_data_2024(self):
        # Load data from noaa
        del_data('*')
        kp_vars = pyspedas.projects.noaa.noaa_load_kp(trange=['2024-01-01', '2024-01-02'])
        self.assertTrue('Kp' in kp_vars)
        self.assertTrue('ap' in kp_vars)
        self.assertTrue('Sol_Rot_Num' in kp_vars)
        self.assertTrue('Sol_Rot_Day' in kp_vars)
        self.assertTrue('Kp_Sum' in kp_vars)
        self.assertTrue('ap_Mean' in kp_vars)
        self.assertTrue('Cp' in kp_vars)
        self.assertTrue('C9' in kp_vars)
        self.assertTrue(data_exists('Kp'))
        self.assertTrue(data_exists('ap'))
        self.assertTrue(data_exists('Sol_Rot_Num'))
        self.assertTrue(data_exists('Sol_Rot_Day'))
        self.assertTrue(data_exists('Kp_Sum'))
        self.assertTrue(data_exists('ap_Mean'))
        self.assertTrue(data_exists('Cp'))
        self.assertTrue(data_exists('C9'))

    def test_load_kp_data_2014(self):
        # Load data from noaa
        del_data('*')
        kp_vars = pyspedas.projects.noaa.noaa_load_kp(trange=['2014-01-01', '2014-01-02'], prefix='noaa_', suffix='_test')
        self.assertTrue('noaa_Kp_test' in kp_vars)
        self.assertTrue('noaa_ap_test' in kp_vars)
        self.assertTrue('noaa_Sol_Rot_Num_test' in kp_vars)
        self.assertTrue('noaa_Sol_Rot_Day_test' in kp_vars)
        self.assertTrue('noaa_Kp_Sum_test' in kp_vars)
        self.assertTrue('noaa_ap_Mean_test' in kp_vars)
        self.assertTrue('noaa_Cp_test' in kp_vars)
        self.assertTrue('noaa_C9_test' in kp_vars)
        self.assertTrue('noaa_Sunspot_Number_test' in kp_vars)
        self.assertTrue('noaa_F10.7_test' in kp_vars)
        self.assertTrue('noaa_Flux_Qualifier_test' in kp_vars)
        self.assertTrue(data_exists('noaa_Kp_test'))
        self.assertTrue(data_exists('noaa_ap_test'))
        self.assertTrue(data_exists('noaa_Sol_Rot_Num_test'))
        self.assertTrue(data_exists('noaa_Sol_Rot_Day_test'))
        self.assertTrue(data_exists('noaa_Kp_Sum_test'))
        self.assertTrue(data_exists('noaa_ap_Mean_test'))
        self.assertTrue(data_exists('noaa_Cp_test'))
        self.assertTrue(data_exists('noaa_C9_test'))
        self.assertTrue(data_exists('noaa_Sunspot_Number_test'))
        self.assertTrue(data_exists('noaa_F10.7_test'))
        self.assertTrue(data_exists('noaa_Flux_Qualifier_test'))

    def test_load_kp_data_2014_gfz(self):
        # final
        del_data('*')
        kp_vars = pyspedas.projects.noaa.noaa_load_kp(trange=['2014-01-01', '2014-01-02'], gfz=True)
        self.assertTrue('Kp' in kp_vars)
        self.assertTrue('ap' in kp_vars)
        self.assertTrue('Sol_Rot_Num' in kp_vars)
        self.assertTrue('Sol_Rot_Day' in kp_vars)
        self.assertTrue('Kp_Sum' in kp_vars)
        self.assertTrue('ap_Mean' in kp_vars)
        self.assertTrue('Cp' in kp_vars)
        self.assertTrue('C9' in kp_vars)
        self.assertTrue(data_exists('Kp'))
        self.assertTrue(data_exists('ap'))
        self.assertTrue(data_exists('Sol_Rot_Num'))
        self.assertTrue(data_exists('Sol_Rot_Day'))
        self.assertTrue(data_exists('Kp_Sum'))
        self.assertTrue(data_exists('ap_Mean'))
        self.assertTrue(data_exists('Cp'))
        self.assertTrue(data_exists('C9'))

if __name__ == "__main__":
    unittest.main()



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pyspedas

def tplot_ascii(tvar, filename=None, extension='.csv'):
    """
    Save a single tplot variable in CSV format.

    This routine converts the internal tplot data structure into a Pandas data frame, then uses
    the Pandas to_csv() function to save it in CSV format.

    If the tplot variable includes spectral bin metadata, it will be written to a separate file with
    a "_v" inserted after the filename and before the ".csv" suffix.

    Parameters
    ----------
        tvar: str
            tplot variable name
        filename: str
            Base filename to use.  If spec_bin data is present, it will be written to a separate file
            with "_v" appended to the base filename. The file extension suffix should not be included here (it will be added internally).
        extension: str
            File extension suffix to apply to the base filename.  Default: ".csv" (no other file formats
            are supported).

    Examples
    --------

    >>> import pyspedas
    >>> pyspedas.projects.themis.state(probe='a', trange=['2007-03-23', '2007-03-24'])
    >>> pyspedas.tplot_ascii('tha_pos',filename='themis_a_position')
    """

    # grab data, prepend index column
    if filename is None:
        filename = tvar
    # save data
    pyspedas.tplot_tools.data_quants[tvar].to_pandas().to_csv(filename + extension)
    # only try to save spec_bins (y-values in spectrograms) if we're sure they exist
    if 'spec_bins' in pyspedas.tplot_tools.data_quants[tvar].coords.keys():
        pyspedas.tplot_tools.data_quants[tvar].coords['spec_bins'].to_pandas().to_csv(filename + '_v' + extension)




import numpy as np
from pyspedas.tplot_tools import get_data, store_data

def xyz_to_polar(data, co_latitude=False):
    """
    Convert cartesian coordinates to polar coordinates.

    Parameters
    ----------
    x : numpy.ndarray
        x-component of the vector
    y : numpy.ndarray
        y-component of the vector
    z : numpy.ndarray
        z-component of the vector

    Returns
    -------
    r : numpy.ndarray
        radial component of the vector
    theta : numpy.ndarray
        polar angle of the vector
    phi : numpy.ndarray
        azimuthal angle of the vector
    """

    make_tvars=False
    if isinstance(data, str):
        make_tvars = True
        name_in = data
        d = get_data(name_in)
        data = d.y

    x = data[:, 0]
    y = data[:, 1]
    z = data[:, 2]
    out = np.zeros(data.shape)
    out[:, 0] = np.sqrt(x**2 + y**2 + z**2)
    out[:, 1] = np.arccos(z/out[:, 0])*180.0/np.pi
    out[:, 2] = np.arctan2(y, x)*180.0/np.pi
    if not co_latitude:
        out[:, 1] = 90.0 - out[:, 1]

    if make_tvars:
       store_data(name_in+'_mag', data={'x':d.times, 'y':out[:,0]})
       store_data(name_in + '_th', data={'x': d.times, 'y': out[:, 1]})
       store_data(name_in + '_phi', data={'x': d.times, 'y': out[:, 2]})
       return
    else:
        return out



from .vhm import vhm
from .swoops import swoops
from .swics import swics
from .urap import urap
from .epac import epac
from .hiscale import hiscale
from .cospin import cospin
from .grb import grb
from .datasets import datasets



import pyspedas
from pyspedas.tplot_tools import store_data, tplot_wildcard_expand
import pandas as pd
import copy
import xarray as xr
import logging
import numpy as np

from pyspedas.tplot_tools import convert_tplotxarray_to_pandas_dataframe


# JOIN TVARS
# join TVars into single TVar with multiple columns
def join_vec(tvars, newname=None, merge=False):
    """
    Joins 1D tplot variables into one tplot variable.

    .. note::
        This analysis routine assumes the data is no more than 2 dimensions. If there are more, they may become flattened!

    Parameters
    ----------
    tvars : list of str
        Name of tplot variables to join together.
    newname : str, optional
        The name of the new tplot variable. If not specified (the default), a name will be assigned.
    merge : bool, optional
        Whether or not to merge the created variable into an older variable.
        Default is False.

    Returns
    -------
    None

    Examples
    --------
    >>> import pyspedas
    >>> import numpy as np
    >>> pyspedas.store_data('d', data={'x':[2,5,8,11,14,17,21], 'y':[[1,1,50],[2,2,3],[100,4,47],[4,90,5],[5,5,99],[6,6,25],[7,7,-5]]})
    >>> pyspedas.store_data('e', data={'x':[2,5,8,11,14,17,21], 'y':[[np.nan,1,1],[np.nan,2,3],[4,np.nan,47],[4,np.nan,5],[5,5,99],[6,6,25],[7,np.nan,-5]]})
    >>> pyspedas.store_data('g', data={'x':[0,4,8,12,16,19,21], 'y':[[8,1,1],[100,2,3],[4,2,47],[4,39,5],[5,5,99],[6,6,25],[7,-2,-5]]})
    >>> pyspedas.join_vec(['d','e','g'],newname='deg')

    """

    to_merge = False
    if newname in pyspedas.tplot_tools.data_quants.keys() and merge:
        prev_data_quant = pyspedas.tplot_tools.data_quants[newname]
        to_merge = True

    # Allow wildcard specification
    tvars=tplot_wildcard_expand(tvars)

    if newname is None:
        newname = "-".join(tvars) + "_joined"

    has_extra_v_values = False
    v_shape = None
    data_shape = None
    combined_v = None

    # Make sure all the components exist and are compatible to be joined
    for i, val in enumerate(tvars):
        dq = pyspedas.tplot_tools.data_quants[val]
        data_shape = dq.values.shape
        if len(data_shape) != 1:
            logging.error(f"join_vec: variable {val} must be 1-dimensional, but has shape {data_shape}.")
            return None
        if i == 0:
            if 'extra_v_values' in dq.attrs.keys():
                has_extra_v_values = True
                extra_v_values = dq.attrs['extra_v_values']
                v_shape = extra_v_values.shape
                # If extra_v_values has multiple elements, maybe it's time varying....it should match the number of timestamps.
                # We'll create the array here, and populate each row as we go.
                if has_extra_v_values:
                    if len(extra_v_values) == 1:
                        combined_v = np.empty((len(tvars),),dtype=extra_v_values.dtype)
                        combined_v[0] = extra_v_values[0]
                    elif len(extra_v_values) == data_shape[0]:
                        combined_v = np.empty((data_shape[0],len(tvars)), dtype=extra_v_values.dtype)
                        combined_v[:,0] = extra_v_values
                    else:
                        logging.error(f"join_vec: variable {val} has shape {data_shape}, but has incompatible extra_v_values shape {v_shape}.")
                        return None
            df = convert_tplotxarray_to_pandas_dataframe(tvars[i], no_spec_bins=True)

        else:
            if 'extra_v_values' in dq.attrs.keys():
                current_has_extra_v_values = True
                current_extra_v_values = dq.attrs['extra_v_values']
                current_v_shape = current_extra_v_values.shape
            else:
                current_has_extra_v_values = False
                current_extra_v_values = None
                current_v_shape = None
            current_data_shape = dq.values.shape

            if current_has_extra_v_values != has_extra_v_values:
                logging.error(f"join_vec: Unable to combine {tvars[0]} and {tvars[i]}, mismatched extra_v_values attributes.")
                return None
            elif has_extra_v_values and v_shape != current_v_shape:
                logging.error(f"join_vec: Unable to combine {tvars[0]} and {tvars[i]}, mismatched extra_v_values shapes {v_shape} and {current_v_shape}.")
                return None
            elif data_shape != current_data_shape:
                logging.error(f"join_vec: Unable to combine {tvars[0]} and {tvars[i]}, mismatched data shapes {data_shape} and {current_data_shape}")
                return None

            if has_extra_v_values and v_shape[0] == 1:
                combined_v[i] = current_extra_v_values[0]
            elif has_extra_v_values and v_shape[0] > 1:
                combined_v[:,i] = current_extra_v_values
            d = convert_tplotxarray_to_pandas_dataframe(tvars[i], no_spec_bins=True)
            df = pd.concat([df, d], axis=1)

    if combined_v is None:
        store_data(newname, data={"x": df.index, "y": df.values})
    else:
        store_data(newname, data={"x": df.index, "y": df.values, "v": combined_v})

    if to_merge is True:
        cur_data_quant = pyspedas.tplot_tools.data_quants[newname]
        plot_options = copy.deepcopy(pyspedas.tplot_tools.data_quants[newname].attrs)
        pyspedas.tplot_tools.data_quants[newname] = xr.concat(
            [prev_data_quant, cur_data_quant], dim="time"
        ).sortby("time")
        pyspedas.tplot_tools.data_quants[newname].attrs = plot_options

    return newname




from pyspedas import options, ylim, get_data

from ..load import load
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional

def mepe(
    trange: List[str] = ['2017-03-27', '2017-03-28'],
    datatype: str = 'omniflux',
    level: str = 'l2',
    suffix: str = '',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the MEP-e experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Dafault: ['2017-03-27', '2017-03-28']
            '
        datatype: str
            Data type; Valid 'l2' options: '3dflux', 'omniflux'  Valid 'l3' options: '3dflux', 'pa'
            Default: 'omniflux'

        level: str
            Data level; Valid options: 'l2', 'l3'
            Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix.  Default: None

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded.  Default: [] (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables.
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword.
            Default: False

        ror: bool
            If set, print PI info and rules of the road.
            Default: True

        uname: str
            User name. Default: None

        passwd: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> mepe_vars = pyspedas.projects.erg.mepe(trange=['2017-03-27', '2017-03-28'])
    >>> tplot('erg_mepe_l2_omniflux_FEDO')

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    if level == 'l3' and datatype not in ['3dflux', 'pa']:
        datatype = '3dflux'

    file_res = 3600. * 24
    prefix = 'erg_mepe_'+level + '_' + datatype + '_'
    pathformat = 'satellite/erg/mepe/'+level+'/'+datatype + \
        '/%Y/%m/erg_mepe_'+level+'_'+datatype+'_%Y%m%d_v??_??.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)

    mixed_case = False
    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)

            # --- print PI info and rules of the road

            print(' ')
            print(
                '**************************************************************************')
            if "LOGICAL_SOURCE_DESCRIPTION" in gatt.keys():
                print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
                mixed_case = False
            elif "Logical_source_description" in gatt.keys():
                if "LOGICAL_SOURCE_DESCRIPTION" in gatt.keys():
                    print(gatt["Logical_source_description"])
                mixed_case = True

            if mixed_case:
                pi_name_att = "PI_name"
                pi_affiliation_att = "PI_affiliation"
            else:
                pi_name_att = "PI_NAME"
                pi_affiliation_att = "PI_AFFILIATION"

            print('')
            print('PI: ', gatt[pi_name_att])
            print("Affiliation: ", gatt[pi_affiliation_att])
            print('')
            print('- The rules of the road (RoR) common to the ERG project:')
            print(
                '      https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                '- RoR for MEP-e data:  https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Mepe')
            print('')
            print('Contact: erg_mep_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if datatype == 'omniflux':
        # set spectrogram plot option
        options(prefix + 'FEDO' + suffix, 'Spec', 1)
        # set y axis to logscale
        options(prefix + 'FEDO' + suffix, 'ylog', 1)
        # set ytitle
        options(prefix + 'FEDO' + suffix, 'ytitle', 'ERG\nMEP-e\nFEDO\nEnergy')
        # set ysubtitle
        options(prefix + 'FEDO' + suffix, 'ysubtitle', '[keV]')
        # set ylim
        ylim(prefix + 'FEDO' + suffix, 6., 100.)
        # set z axis to logscale
        options(prefix + 'FEDO' + suffix, 'zlog', 1)
        # set ztitle
        options(prefix + 'FEDO' + suffix, 'ztitle', '[/s-cm^{2}-sr-keV]')
        # change colormap option
        options(prefix + 'FEDO' + suffix, 'Colormap', 'jet')
    elif (datatype == '3dflux') and (level == 'l2'):
        # set spectrogram plot option
        options(prefix + 'FEDU' + suffix, 'Spec', 1)
        options(prefix + 'FEDU_n' + suffix, 'Spec', 1)
        options(prefix + 'FEEDU' + suffix, 'Spec', 1)
        options(prefix + 'count_raw' + suffix, 'Spec', 1)

        # set y axis to logscale
        options(prefix + 'FEDU' + suffix, 'ylog', 1)
        options(prefix + 'FEDU_n' + suffix, 'ylog', 1)
        options(prefix + 'FEEDU' + suffix, 'ylog', 1)
        options(prefix + 'count_raw' + suffix, 'ylog', 1)

        # set ysubtitle
        options(prefix + 'FEDU' + suffix, 'ysubtitle', '[keV]')
        options(prefix + 'FEDU_n' + suffix, 'ysubtitle', '[keV]')
        options(prefix + 'count_raw' + suffix, 'ysubtitle', '[keV]')

        # set ylim
        ylim(prefix + 'FEDU' + suffix, 6., 100.)
        ylim(prefix + 'FEDU_n' + suffix, 6., 100.)
        ylim(prefix + 'count_raw' + suffix, 6., 100.)

        # set z axis to logscale
        options(prefix + 'FEDU' + suffix, 'zlog', 1)
        options(prefix + 'FEDU_n' + suffix, 'zlog', 1)
        options(prefix + 'FEEDU' + suffix, 'zlog', 1)
        options(prefix + 'count_raw' + suffix, 'zlog', 1)

        # set ztitle
        options(prefix + 'FEDU' + suffix, 'ztitle', '[/s-cm^{2}-sr-keV]')
        options(prefix + 'FEDU_n' + suffix, 'ztitle', '[/s-cm^{2}-sr-keV]')

        # change colormap option
        options(prefix + 'FEDU' + suffix, 'Colormap', 'jet')
        options(prefix + 'FEDU_n' + suffix, 'Colormap', 'jet')
        options(prefix + 'FEEDU' + suffix, 'Colormap', 'jet')
        options(prefix + 'count_raw' + suffix, 'Colormap', 'jet')

    return loaded_data



import numpy as np
import logging
from .lingradest import lingradest
from pyspedas import tinterpol
from pyspedas.tplot_tools import get_data, store_data, time_double, time_string, time_clip, deflag
from pyspedas.tplot_tools import tnames, options, tplot_options, tsmooth, tplot, get_coords, get_units, set_coords, set_units

def classify_null_type(lambdas_in):
    """
    Determine the topological type of a magnetic null, given the eigenvalues of the Jacobian matrix.

    Parameters
    ----------
    lambdas_in: An array of 3 complex-valued eigenvalues of the Jacobian matrix,
        in no particular order.

    Returns
    -------
    int
        An integer representing the null type::

       0: Unknown/unable to characterize
       1: X-type null
       2: O-type null (island or plasmoid)
       3: A-type (radial) null
       4: B-type (radial) null
       5: A_s-type (spiral) null
       6: B_s-type (spiral) null
       7: X-type, degenerated from A type
       8: X-type, degenerated from B type
       9: O-type, degenerated from A_s type
       10: O-type, degenerated from B_s type

    See Table 1, Fu et al 2015 for details

    References
    ----------
    Fu, H. S., A. Vaivads, Y. V. Khotyaintsev, V. Olshevsky, M. André, J. B. Cao, S. Y. Huang,
    A. Retinò, and G. Lapenta (2015), How to find magnetic nulls and reconstruct field topology
    with MMS data?. J. Geophys. Res. Space Physics, 120, 3758–3782. doi: 10.1002/2015JA021082.

    Paschmann, G., Daly, P. (1998), Analysis Methods for Multi-Spacecraft Data, ISSR
    """

    lambda1 = lambdas_in[0]
    lambda2 = lambdas_in[1]
    lambda3 = lambdas_in[2]

    # We want to find the max and min (absolute value) norms, real, and imaginary parts,
    # to decide later if 3-d type nulls should degenerate to 2-d type nulls

    n1 = np.abs(lambda1)
    n2 = np.abs(lambda2)
    n3 = np.abs(lambda3)
    nlist = np.array([n1,n3,n3])
    n_max = np.max(nlist)
    n_min = np.min(nlist)

    re1 = np.abs(lambda1.real)
    re2 = np.abs(lambda2.real)
    re3 = np.abs(lambda2.real)
    re_list = np.array([re1,re2,re3])
    re_max = np.max(re_list)
    re_min = np.min(re_list)

    im1 = np.abs(lambda1.imag)
    im2 = np.abs(lambda2.imag)
    im3 = np.abs(lambda3.imag)
    # One of the eigenvalues is always pure real. So for this comparison, we only want to
    # consider the ones that might have imaginary components.  If they're all real, it won't matter.
    if (im1 == 0.0):
        im_list = np.array([im2,im3])
    elif (im2 == 0.0):
        im_list = np.array([im1,im3])
    else:
        im_list = np.array([im1,im2])
    im_max = np.max(im_list)
    im_min = np.min(im_list)


    # Now we order by size of the norms
    if (n1 <= n2) and (n2 <= n3):
        s1 = lambda1
        s2 = lambda2
        s3 = lambda3
    elif (n1 <= n3) and (n3 <= n2):
        s1 = lambda1
        s2 = lambda3
        s3 = lambda2
    elif (n2 <= n3) and (n3 <= n1):
        s1 = lambda2
        s2 = lambda3
        s3 = lambda1
    elif (n2 <= n1) and (n1 <= n3):
        s1 = lambda2
        s2 = lambda1
        s3 = lambda3
    elif (n3 <= n1) and (n1 <= n2):
        s1 = lambda3
        s2 = lambda1
        s3 = lambda2
    else:
        s1 = lambda3
        s2 = lambda2
        s3 = lambda1

    typecode = 0  # Default to unclassified

    # Now we can classify the null type.  It should be safe to use floating point equality to check
    # the imaginary parts.  It is unlikely that any of the real parts or norms will be exactly 0,
    # but see below where we check for some possibly degenerate cases.

    # Note: this logic appears to classify everything, with no "unknown" type nulls.  But in
    # Hu's paper, unknown type seems to happen fairly frequently. I suspect some of my typecode 5 and
    # 6 classifications might become "unknown" if I also looked at the components of the complex eigenvalues,
    # rather than just checking the sign of the pure real.

    if (s1.imag == 0.0) and (s2.imag == 0.0) and (s3.imag == 0.0):  # all eigenvalues pure real
        if (s1.real == 0.0):  # smallest (absolute val) real is exactly 0
            typecode = 1 # type X
        elif (s3.real > 0.0): # largest (absolute val) real part is +ve
            typecode = 3 # type A
        else: # largest (abs) real part is -ve
            typecode = 4  # type B
    elif (n1 == 0.0): # one pure real == 0, other two imaginary
        typecode = 2 # O-type
    else:  # all eigenvalues nonzero, one pure real two imaginary
        if (s1.imag == 0.0): # s1 is the pure real eigenvalue
            if s1.real > 0.0:   # real eigenvalue is +ve
                typecode = 5 # A_s
            else:               # real eigenvalue is -ve
                typecode = 6 # B_s
        elif (s2.imag == 0.0): # s2 is the pure real eigenvalue
            if s2.real > 0.0:   # real eigenvalue is +ve
                typecode = 5 # A_s
            else:               # real eigenvalue is -ve
                typecode = 6 # B_s
        else:   # s3 must be the pure real
            if s3.real > 0.0:
                typecode = 5 # A_s
            else:
                typecode = 6 # B_s


    # Now check to see if a 3-d null should degenerate to a 2-d type of null
    # Type A and B nulls should degenerate to type X if min(norms) < .25*max(norms)
    # Type A_s and B_s nulls should degenerate to type O if max(real) < 0.25*min(imag)

    if (typecode == 3) or (typecode == 4): # Type A or B
        if n_min < 0.25 * n_max:
            typecode = typecode + 4 # degenerate to X-type
    elif (typecode == 5) or (typecode == 6):  # Type A_s or B_s
        if re_max < 0.25*im_min:
            typecode = typecode + 4   # degenerate to O-type
    return typecode


# Define some stuff we'll use in debugging messages and plot options

typecode_strings = ['Unknown','X','O','A','B','A_s', 'B_s','X_a','X_b','O_a','O_b']
typecode_symbols = ['.','x','$o$','4','4','>','>','x','x','$o$','$o$']
typecode_colors = ['k','k','k','r','b','r','b','r','b','r','b']

def find_magnetic_nulls_fote(positions=None, fields=None, smooth_fields=True, smooth_npts=10, smooth_median=True,scale_factor=1.0):
    """
    Find magnetic null points, using the First Order Taylor Expansion (FOTE) method, from a set of four-point magnetic field observations.

    Parameters
    -----------

    positions: list of str
        A 4-element list of tplot variable names representing the probe positions

    fields: list of str
        A 4-element list of tplot variable names representing the magnetic field measurements

    smooth_fields: bool
        If True, perform boxcar averaging on the fields
        Default: True

    smooth_npts: int
        Number of points to use in the boxcar smoothing, if smoothing is enabled.

    smooth_median: bool
        If True and smoothing is enabled, use median filtering
        Default: True

    scale_factor: float
        The scale factor passed to the lingradest routine to scale some of the distances
        Default: 1.0

    Returns
    -------
    list of str
        A list of tplot variables describing the nulls found::

            'null_pos': Position of the null point, in the same coordinate system as imput positions
            'null_bary_dist': Distance between the null point and the barycenter of the tetrahedron
            'null_bary_dist_types': A composite variable more suitable for plotting, with the null to barycenter distances,
                  superimposed with symbols representing the type of each null found
            'null_sc_distances': The distances from the null to each of the four spacecraft
            'null_fom': Figures of merit 'eta' and 'xi', roughly representing the confidence in the null location and null type.
                  Lower is better, with values less than 0.4 denoting fairly reliable detection and classification
            'null_typecode': The type of each null point found, with values from 0-6. See classify_null_type() for interpretation.
            'max_reconstruction_error': The maximum error out of the four s/c, when using the calculated Jacobian and field
                  at the barycenter to reconstruct the field vectors at each spacecraft.  Should be extremely close to zero.


    Notes
    ------

    This routine uses tetrahedral interpolation to estimate the magnetic field (B0) and field gradients at the tetrahedron
    barycenter. From the field gradients we can construct a Jacobian matrix (J).  The field in that region can be
    expressed, to a first order approximation, as

    B = B0 + JV

    where V is the position vector relative to the barycenter.  At a magnetic null V_null, all components of B are zero,
    so we have JV_null = -B0, which can be solved to get V_null.

    As long as the four field measurements all differ, a null point will always be found.   For it to be credible,
    it should be in the neighborhood where the linear gradient approximation is expected to be valid, i.e. some smallish multiple
    of the tetrahedron size.

    The topology of the field around the null can be inferred from the eigenvalues and eigenvectors of the
    estimated Jacobian. See the classify_null_type() function for details.

    References
    -----------

    Fu, H. S., A. Vaivads, Y. V. Khotyaintsev, V. Olshevsky, M. André, J. B. Cao, S. Y. Huang,
    A. Retinò, and G. Lapenta (2015), How to find magnetic nulls and reconstruct field topology
    with MMS data?. J. Geophys. Res. Space Physics, 120, 3758–3782. doi: 10.1002/2015JA021082.

    Paschmann, G., Daly, P. (1998), Analysis Methods for Multi-Spacecraft Data, ISSR


    Example
    -------

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> data = pyspedas.projects.mms.fgm(probe=[1, 2, 3, 4], trange=['2015-09-19/07:40', '2015-09-19/07:45'], data_rate='srvy', time_clip=True, varformat='*_gse_*', get_fgm_ephemeris=True)
    >>> fields = ['mms'+prb+'_fgm_b_gse_srvy_l2' for prb in ['1', '2', '3', '4']]
    >>> positions = ['mms'+prb+'_fgm_r_gse_srvy_l2' for prb in ['1', '2', '3', '4']]
    >>> null_vars = pyspedas.find_magnetic_nulls_fote(fields=fields, positions=positions, smooth_fields=True,smooth_npts=10,smooth_median=True)
    >>> tplot(null_vars)
    """

    # Input data needs to be sanitized, by removing any nans, and finding a time range common
    # to all the deflagged variables

    positions_df = []
    fields_df = []
    fields_sm = []

    for i in range(4):
        positions_df.append(positions[i]+'_df')
        fields_df.append(fields[i]+'_df')
        if smooth_fields:
            fields_sm.append(fields[i]+'_sm')
        else:
            fields_sm.append(fields[i]+'_df')

        deflag(positions[i], method='remove_nan', newname=positions_df[i])
        deflag(fields[i], method='remove_nan', newname=fields_df[i])
        if smooth_fields:
            tsmooth(fields_df[i], width=smooth_npts, median=smooth_median, newname=fields_sm[i])

    # Interpolate all deflagged positions and field measurements to the first probe's field measurement times

    # We need to find a time range that's contained in all the position and field variables, to
    # prevent interpolation from introducing NaNs and causing exceptions to be thrown in the calculations
    # below.

    tmin = 0.0
    tmax = time_double('3000-01-01')
    for i in range(4):
        d = get_data(positions_df[i])
        dmin = d.times[0]
        dmax = d.times[-1]
        tmin = np.max([tmin, dmin])
        tmax = np.min([tmax, dmax])

        d = get_data(fields_sm[i])
        dmin = d.times[0]
        dmax = d.times[-1]
        tmin = np.max([tmin, dmin])
        tmax = np.min([tmax, dmax])

    # Shave another millisecond off each end just to be sure (because floating point equality testing is weird)
    tmin = tmin + .001
    tmax = tmax - .001

    # We'll interpolate everything to the first set of times, so we'll time_clip that one to the desired time range
    time_clip(fields_sm[0], tmin, tmax, newname='f0_tc')

    # Now we can interpolate safely
    tinterpol(positions_df,'f0_tc',newname=['pos1_i','pos2_i','pos3_i','pos4_i'])
    tinterpol(fields_sm, 'f0_tc', newname=['b1_i','b2_i','b3_i','b4_i'])

    # Get the data arrays from the interpolated tplot variables
    d1 = get_data('pos1_i')
    d2 = get_data('pos2_i')
    d3 = get_data('pos3_i')
    d4 = get_data('pos4_i')

    # Position variables for MMS include a total distance as the last element, so only take the first three
    r1=d1.y[:,0:3]
    r2=d2.y[:,0:3]
    r3=d3.y[:,0:3]
    r4=d4.y[:,0:3]

    b1 = get_data('b1_i')
    b2 = get_data('b2_i')
    b3 = get_data('b3_i')
    b4 = get_data('b4_i')

    # The MMS field variables also include a fourth component with the total field
    bx1 = b1.y[:,0]
    by1 = b1.y[:,1]
    bz1 = b1.y[:,2]

    bx2 = b2.y[:,0]
    by2 = b2.y[:,1]
    bz2 = b2.y[:,2]

    bx3 = b3.y[:,0]
    by3 = b3.y[:,1]
    bz3 = b3.y[:,2]

    bx4 = b4.y[:,0]
    by4 = b4.y[:,1]
    bz4 = b4.y[:,2]

    datapoint_count = len(bx1)
    times = b1.times


    # Call the lingradest routine to get the barycenter locations, field at barycenter, and field gradients at barycenter

    # Output of lingradest is a dictionary of string keys and numpy array values

    lingrad_output = lingradest(bx1, bx2, bx3, bx4,
                                by1, by2, by3, by4,
                                bz1, bz2, bz3, bz4,
                                r1,r2,r3,r4, scale_factor=scale_factor)

    # Output arrays
    out_pos_null = np.zeros((datapoint_count,3))
    out_pos_bary = np.zeros((datapoint_count,3))
    out_null_bary_dist = np.zeros((datapoint_count))
    out_null_p1_dist = np.zeros((datapoint_count))
    out_null_p2_dist = np.zeros((datapoint_count))
    out_null_p3_dist = np.zeros((datapoint_count))
    out_null_p4_dist = np.zeros((datapoint_count))
    out_null_eta = np.zeros((datapoint_count))
    out_null_xi = np.zeros((datapoint_count))
    out_typecode = np.zeros((datapoint_count))
    out_max_reconstruction_error = np.zeros((datapoint_count))

    for i in range(datapoint_count):
        Rbary = lingrad_output['Rbary'][i]
        out_pos_bary[i,:] = Rbary[:]

        # Get field vector at barycenter for i-th point
        bxbc = lingrad_output['Bxbc'][i]
        bybc = lingrad_output['Bybc'][i]
        bzbc = lingrad_output['Bzbc'][i]
        b0 = np.array((bxbc, bybc, bzbc)) # Field at barycenter
        b0_neg = np.array((-bxbc,-bybc,-bzbc))  # Negative of field at barycenter

        # Form Jacobian matrix from field gradients at barycenter
        LGBx = lingrad_output['LGBx'][i]
        LGBy = lingrad_output['LGBy'][i]
        LGBz = lingrad_output['LGBz'][i]
        J = np.stack((LGBx, LGBy, LGBz))  # Jacobian matrix

        # Solve for null position with respect to barycenter (maybe check for singular matrix first?)

        r_null = np.linalg.solve(J,b0_neg)
        F_null = b0 + np.matmul(J,r_null)
        r_null_bary_dist = np.linalg.norm(r_null)
        out_null_bary_dist[i] = r_null_bary_dist

        # Translate r_null to origin of probe coordinate system
        pos_null = r_null + Rbary
        out_pos_null[i,:] = pos_null[:]

        # Get distances from null to each probe in the tetrahedron
        p1_null = r1[i] - pos_null
        p1_null_dist = np.linalg.norm(p1_null)
        p2_null = r2[i] - pos_null
        p2_null_dist = np.linalg.norm(p2_null)
        p3_null = r3[i] + pos_null
        p3_null_dist = np.linalg.norm(p3_null)
        p4_null = r4[i] - pos_null
        p4_null_dist = np.linalg.norm(p4_null)
        min_null_dist = np.min(np.array([p1_null_dist,p2_null_dist,p3_null_dist,p4_null_dist]))

        out_null_p1_dist[i] = p1_null_dist
        out_null_p2_dist[i] = p2_null_dist
        out_null_p3_dist[i] = p3_null_dist
        out_null_p4_dist[i] = p4_null_dist

        # Estimate field at each probe using the estimated linear gradient
        # These should all be quite close to the measured fields.

        dR1 = -1.0 * lingrad_output['dR1'][i]
        dR2 = -1.0 * lingrad_output['dR2'][i]
        dR3 = -1.0 * lingrad_output['dR3'][i]
        dR4 = -1.0 * lingrad_output['dR4'][i]

        F1_est = b0 + np.matmul(J,dR1)
        F2_est = b0 + np.matmul(J,dR2)
        F3_est = b0 + np.matmul(J,dR3)
        F4_est = b0 + np.matmul(J,dR4)

        F1_obs = np.array([bx1[i],by1[i],bz1[i]])
        F2_obs = np.array([bx2[i],by2[i],bz2[i]])
        F3_obs = np.array([bx3[i],by3[i],bz3[i]])
        F4_obs = np.array([bx4[i],by4[i],bz4[i]])

        F1_err = F1_est - F1_obs
        F2_err = F2_est - F2_obs
        F3_err = F3_est - F3_obs
        F4_err = F4_est - F4_obs

        F1_err_norm = np.linalg.norm(F1_err)
        F2_err_norm = np.linalg.norm(F2_err)
        F3_err_norm = np.linalg.norm(F3_err)
        F4_err_norm = np.linalg.norm(F4_err)

        max_reconstruction_error = np.max(np.array([F1_err_norm,F2_err_norm,F3_err_norm,F4_err_norm]))
        out_max_reconstruction_error[i] = max_reconstruction_error

        # Look for opposite signs in each field component, necessary if null is within tetrahedron
        bxmax = np.max( (F1_obs[0],F2_obs[0],F3_obs[0],F4_obs[0]))
        bxmin = np.min( (F1_obs[0],F2_obs[0],F3_obs[0],F4_obs[0]))
        bymax = np.max( (F1_obs[1],F2_obs[1],F3_obs[1],F4_obs[1]))
        bymin = np.min( (F1_obs[1],F2_obs[1],F3_obs[1],F4_obs[1]))
        bzmax = np.max( (F1_obs[2],F2_obs[2],F3_obs[2],F4_obs[2]))
        bzmin = np.min( (F1_obs[2],F2_obs[2],F3_obs[2],F4_obs[2]))

        tstring = time_string(times[i])

        # Get eigenvalues of J to characterize field topology near the null
        eigenvalues, eigenvectors = np.linalg.eig(J)
        lambdas = eigenvalues
        l1_norm = np.linalg.norm(lambdas[0])
        l2_norm = np.linalg.norm(lambdas[1])
        l3_norm = np.linalg.norm(lambdas[2])
        eig_sum_norm = np.linalg.norm(lambdas[0] + lambdas[1] + lambdas[2])
        eig_max_norm = np.max((l1_norm,l2_norm,l3_norm))
        # Determine the toplogical type of the null by inspecting the eigenvalues
        typecode = classify_null_type(lambdas)
        out_typecode[i] = typecode

        # div B should be close to 0 (exactly 0 in theory), so the difference from 0 is a measure of how
        # credible any null we've found might be.  We normalize it, dividing by the magnitude of the curl,
        # to get the statistic eta.  eta < 0.40 for a credible null.
        divB = lingrad_output['LD'][i]
        LCx = lingrad_output['LCxB'][i]
        LCy = lingrad_output['LCyB'][i]
        LCz = lingrad_output['LCzB'][i]
        curlB = np.array([LCx,LCy,LCz])
        curlB_norm = np.linalg.norm(curlB)
        # eta - | del dot B| /  |del x B|
        eta = np.abs(divB)/curlB_norm
        out_null_eta[i] = eta

        # Similarly, the sum of the eigenvectors of the Jacobian gives another figure of merit.  Here,
        # we normalize by dividing the norm of the sum by the norm of the largest eigenvalue to yield xi.
        # xi < 0.40 if we trust the topology derived from the eigenvalues.  Larger values may mean that the type of
        # null detected (radial vs. spiral, etc) may be an artifact.

        xi = eig_sum_norm/eig_max_norm
        out_null_xi[i] = xi

        show_in_box = False
        if show_in_box and (bxmax*bxmin < 0.0) and (bymax*bymin < 0.0) and (bzmax*bzmin < 1.0):
            print("Null of type ",typecode_strings[typecode]," may be in tetrahedron at time ",tstring )
            if times[i] > time_double('2003-08-17/16:41:55.43'):
                print("Jacobian:")
                print(LGBx)
                print(LGBy)
                print(LGBz)
                print("Eigenvalues:")
                for l in lambdas: print(l)
                print("Eigenvectors:")
                for e in eigenvectors: print(e)
                print("r_null", r_null)
                print("Max reconstruction error:",max_reconstruction_error)
                print("Distance from null to nearest spacecraft",min_null_dist)
                print("eta:", eta)
                print("xi",xi)


    # Now create output tplot variables and set some plot options

    pos_coords = get_coords(positions[0])
    pos_units = get_units(positions[0])
    field_coords = get_coords(fields[0])
    field_units = get_units(fields[0])
    store_data('null_pos',data={'x':times,'y':out_pos_null})
    set_units('null_pos',pos_units)
    set_coords('null_pos',pos_coords)

    store_data('null_bary_dist',data={'x':times,'y':out_null_bary_dist})
    set_units('null_bary_dist',pos_units)
    set_coords('null_bary_dist',pos_coords)
    options('null_bary_dist','yrange',[0.0, 1000.0])
    options('null_bary_dist','ytitle','Null dist')

    # Create a set of variables with each null type, so we can plot each type with a different symbol
    symvars = ['null_bary_dist']
    symvars_colors = ['k']
    symvars_legends = ['Distance']
    symvars_markers = ["none"]
    all_idx=np.arange(len(out_typecode))
    for i in range(11):
        cond = out_typecode == i
        idx = all_idx[cond]
        if len(idx) > 0:
            tvar = 'null_type'+str(i)
            store_data(tvar,data={'x':times[idx],'y':out_null_bary_dist[idx]})
            options(tvar, 'marker',typecode_symbols[i])
            options(tvar,'markersize',10)
            #options(tvar, 'color',typecode_colors[i])
            options(tvar,'symbols',True)
            symvars_colors.append(typecode_colors[i])
            symvars_legends.append(typecode_strings[i])
            symvars_markers.append(typecode_symbols[i])
            symvars.append(tvar)
    # Make a pseudovariable combining the different typecodes, so we get a line plot from magnull_null_bary_distance,
    # and a scatter plot of the symbols showing the null classification types, all in one panel.
    store_data('null_bary_dist_types',data=symvars)
    options('null_bary_dist_types','color',symvars_colors)
    options('null_bary_dist_types','legend_names',symvars_legends)
    options('null_bary_dist_types','legend_location','upper right')
    options('null_bary_dist_types','legend_ncols',len((symvars)))
    options('null_bary_dist_types','legend_markerfirst',True)
    options('null_bary_dist_types','legend_linewidth', 1)
    options('null_bary_dist_types','ytitle','Null types')
    set_units('null_bary_dist_types',pos_units)

    # Distances from null to each probe
    p_distances = np.stack((out_null_p1_dist,out_null_p2_dist,out_null_p3_dist,out_null_p4_dist),axis=1)
    store_data('null_sc_distances',data={'x':times,'y':p_distances})
    set_units('null_sc_distances',pos_units)
    options('null_sc_distances','color',['k','r','g','b'])
    options('null_sc_distances', 'yrange', [0.0, 1000.0])
    options('null_sc_distances','ytitle','Null-sc dist')
    options('null_sc_distances','legend_names',['P1','P2','P3','P4'])
    options('null_sc_distances','legend_ncols',4)

    store_data('null_eta', data={'x':times, 'y':out_null_eta})
    options('null_eta','yrange',[0.0,2.0])
    options('null_eta','ytitle',r"$\eta$")
    store_data('null_xi', data={'x':times, 'y':out_null_xi})
    options('null_xi','yrange',[0.0,2.0])
    options('null_xi','ytitle',r'$\xi$')

    merits = np.stack((out_null_eta,out_null_xi),axis=1)
    store_data('null_fom', data={'x':times, 'y':merits})
    options('null_fom','color',['k','m'])
    options('null_fom','yrange',[0.0,2.0])
    options('null_fom','ytitle','Quality')
    options('null_fom','legend_names',[r'$\eta$',r'$\xi$'])
    options('null_fom','legend_location','upper right')
    options('null_fom','legend_markerfirst',True)

    #print('Minimum typecode:',np.min(out_typecode))
    store_data('null_typecode',{'x':times,'y':out_typecode})
    options('null_typecode','ytitle','Null types')
    options('null_typecode','yrange',[0.0,11.0])

    store_data('max_reconstruction_error',data={'x':times,'y':out_max_reconstruction_error})
    options('max_reconstruction_error','ytitle','Max error')
    set_units('max_reconstruction_error', field_units)

    return_vars = ['null_pos', 'null_bary_dist', 'null_bary_dist_types', 'null_sc_distances', 'null_fom','null_typecode', 'max_reconstruction_error']

    return return_vars



import numpy as np

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean


def spd_pgs_make_e_spec(data_in):
    """
    Builds energy spectrogram from the particle data structure
    
    Parameters
    ----------
        data_in: dict
            Particle data structure

    Returns
    -------
    tuple
        Tuple containing: (energy values for the y-axis, spectrogram values)

    """

    data = data_in.copy()

    # zero inactive bins to ensure areas with no data are represented as NaN
    zero_bins = np.argwhere(data['bins'] == 0)
    if zero_bins.size != 0:
        for item in zero_bins:
            data['data'][item[0], item[1]] = 0.0

    ave = nanmean(data['data'], axis=1)

    y = data['energy'][:, 0]

    return (y, ave)



"""
Multiplies data by stored spectrogram bins, creates new tplot variable

Notes
-----
Similar to spec_mult.pro in IDL SPEDAS.

"""

import pyspedas
from pyspedas.tplot_tools import store_data, convert_tplotxarray_to_pandas_dataframe
import pandas as pd
import copy
import logging

def spec_mult(
        tvar,
        newname=None,
):

    """
    Multiplies the data by the stored spectrogram bins and created a new tplot variable

    Note::
        This analysis routine assumes the data is no more than 2 dimensions.
        If there are more, they may become flattened!

    Parameters
    ----------
        tvar : str
            Name of tplot variable
        newname : str
            Name of new tvar in which to store interpolated data.
            Default: If none is specified, a name will be created of the form tvar_specmult.

    Returns
    -------
        None

    Examples
    --------
        >>> pyspedas.store_data('h', data={'x':[0,4,8,12,16,19,21], 'y':[[8,1,1],[100,2,3],[4,2,47],[4,39,5],[5,5,99],[6,6,25],[7,-2,-5]],'v':[[1,1,50],[2,2,3],[100,4,47],[4,90,5],[5,5,99],[6,6,25],[7,7,-5]]})
        >>> pyspedas.spec_mult('h')

    """

    if newname is None:
        newname = tvar+'_specmult'
    if 'spec_bins' not in pyspedas.tplot_tools.data_quants[tvar].coords:
        logging.error("Specified variable must have spec bins stored.  Returning...")
        return
    d, s = convert_tplotxarray_to_pandas_dataframe(tvar)
    dataframe = d.values
    specframe = s.values
    new_df = pd.DataFrame(dataframe*specframe, columns=d.columns, index=d.index)
    store_data(newname,data={'x': new_df.index,'y': new_df.values})
    pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)
    return



import numpy as np

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum


def mms_pgs_make_e_spec(data_in):
    """
    Builds an energy spectrogram from a simplified particle data structure.

    Parameters
    ----------
    data_in : dict
        The input data structure.

    Returns
    -------
    outtable : ndarray, shape (ny,)
        The energy bins.
    ave : ndarray, shape (ny,)
        The spectrogram.

    Notes
    -----
    - Each energy bin in the output spectrogram (`ave`) is the weighted average
      of the corresponding bins in the input data (`data_in`).
    - The input data is sanitized by zeroing inactive bins to ensure areas with
      no data are represented as NaN.
    - The function uses the first energy table for rebinning the data.
    """
    data = data_in.copy()

    # zero inactive bins to ensure areas with no data are represented as NaN
    zero_bins = np.argwhere(data['bins'] == 0)
    if zero_bins.size != 0:
        for item in zero_bins:
            data['data'][item[0], item[1]] = 0.0

    # use the original energy table for now
    outtable = data['orig_energy']
    outbins = np.zeros([len(data['data'][:, 0]), len(data['data'][0, :])])

    # energy range to accept (this is questionable, but matches IDL)
    # TODO: Take a closer look at this logic.  Should we be comparing to the energy bin _center_ values?
    # The IDL version accepts an 'energy' parameter to define energy limits; that feature may no longer be used though?

    erange = [np.min(data['energy']), np.max(data['energy'])]

    # rebin the data to the original energy table
    for ang_idx in range(0, len(data['data'][0, :])):
        etable = data['energy'][:, ang_idx]
        for binidx in range(0, len(outtable)):
            if data['data'][binidx, ang_idx] != 0.0:
                this_en = find_nearest_neighbor(outtable, [etable[binidx]])
                whereen = np.argwhere(outtable == this_en)
                outbins[whereen, ang_idx] += data['data'][binidx, ang_idx]

    # check for out of range values (questionable but matches IDL)
    where_out_of_erange = np.argwhere((outtable < erange[0]) | (outtable > erange[1]))
    if where_out_of_erange.size != 0:
        outbins[where_out_of_erange] = np.nan

    # If we want to use NaN as a marker for unused bins, we need to use sum rather than nansum here.

    if len(data['data'][0, :]) > 1:
        ave = np.sum(outbins, axis=1)/np.sum(data['bins'], axis=1)
    else:
        ave = outbins/data['bins']

    return outtable, ave


def find_nearest_neighbor(table, item):
    table = np.array(table)
    return min(table, key=lambda p: sum((p - item)**2))



import logging
import numpy as np
from pyspedas.tplot_tools import time_string, time_double
from datetime import datetime, timezone


def dailynames(directory='',
               trange=None,
               res=24*3600.,
               hour_res=False,
               file_format='%Y%m%d',
               prefix='',
               suffix='',
               return_times=False):
    """
    Creates a list of file names using a time range, resoution and file format
    Based on Davin Larson's file_dailynames in IDL SPEDAS

    Parameters:
        directory: str
            String containing the directory for the generated file names
            Default: ''. If directory is not specified the names will be written
            to the current working directory

        trange: list of str, list of datetime or list of floats
            Two-element list containing the start and end times for the file names
            Default: Current date. If the date range is not specified the routine will
            use the current date.

        res: float
            File name resolution in seconds
            Default: 24*3600., i.e., daily)

        file_format: str
            Format of the file names using strftime directives; for reference: https://strftime.org
            Default: %Y%m%d, i.e., daily

        prefix: str
            file name prefix
            Default: ''

        suffix: str
            file name suffix
            Default: ''

    Returns:
        List containing filenames

    Example:
        >>> from pyspedsas import time_float
        >>> time_range = time_float(['2021-01-01','2021-01-03'])
        >>> pyspedas.dailynames(trange=time_range)
    """
    if trange is None:
        logging.error('No trange specified')
        return

    if hour_res:
        res = 3600.
        file_format = '%Y%m%d%H'

    # allows the user to pass in trange as list of datetime objects
    if type(trange[0]) == datetime and type(trange[1]) == datetime:
        trange = [time_string(trange[0].replace(tzinfo=timezone.utc).timestamp()),
                  time_string(trange[1].replace(tzinfo=timezone.utc).timestamp())]

    tr = [trange[0], trange[1]]
    
    if isinstance(trange[0], str):
        tr[0] = time_double(trange[0])
    if isinstance(trange[1], str):
        tr[1] = time_double(trange[1])

    # Davin's magic heisted from file_dailynames in IDL
    mmtr = [np.floor(tr[0]/res), np.ceil(tr[1]/res)]

    if mmtr[1]-mmtr[0] < 1:
        n = 1
    else:
        n = int(mmtr[1]-mmtr[0])

    times = [(float(num)+mmtr[0])*res for num in range(n)]

    if return_times:
        return times

    dates = []
    files = []
        
    for time in times:
        if time_string(time, fmt=file_format) not in dates:
            dates.append(time_string(time, fmt=file_format))
            
    for date in dates:
        files.append(directory + prefix + date + suffix)

    return files

def yearlynames(
    trange,
    file_format="%Y%m%d",
    resolution="year",
    directory="",
    prefix="",
    suffix="",
):
    """
    Creates a list of file names using a time range, resolution, and file format.

    Parameters:
        trange: list of str or list of datetime
            Two-element list containing the start and end times for the file names.

        resolution: str
            Either 'year' or 'half-year' to determine the resolution.

        file_format: str
            Format of the file names using strftime directives.

        directory: str
            String containing the directory for the generated file names.

        prefix: str
            file name prefix.

        suffix: str
            file name suffix.

    Returns:
        List containing filenames.
    """

    if trange is None:
        raise ValueError("No trange specified")

    start_date = (
        trange[0]
        if isinstance(trange[0], datetime)
        else datetime.strptime(trange[0], "%Y-%m-%d")
    )
    end_date = (
        trange[1]
        if isinstance(trange[1], datetime)
        else datetime.strptime(trange[1], "%Y-%m-%d")
    )

    dates = []

    if resolution == "half-year":
        # Adjust the start date to be 6 months prior to the specified start date
        if start_date.month <= 6:
            adjusted_start_date = datetime(start_date.year - 1, start_date.month + 6, 1)
        else:
            adjusted_start_date = datetime(start_date.year, start_date.month - 6, 1)

        # Generate all the January 1st and July 1st dates within the range
        potential_dates = [datetime(year, month, 1) for year in range(start_date.year, end_date.year + 1) for month in [1, 7]]
        
        # Filter out dates outside the range
        dates = [date for date in potential_dates if adjusted_start_date < date <= end_date]
    elif resolution == "year":
        dates = [datetime(year, 1, 1) for year in range(start_date.year, end_date.year + 1)]
    else:
        raise ValueError("Invalid resolution specified. Choose 'year' or 'half-year'.")

    files = [directory + prefix + date.strftime(file_format) + suffix for date in dates]
    return files



from pyspedas import options, ylim, get_data

from ..load import load
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional

def mepi_tof(
    trange: List[str] = ['2017-03-27', '2017-03-28'],
    datatype: str = 'flux',
    level: str = 'l2',
    suffix: str = '',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the MEP-i experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-03-27', '2017-03-28']

        datatype: str
            Data type; Valid options: 'flux'
            Default: 'flux'

        level: str
            Data level; Valid options: 'l2'
            Defauls: 'l2'

        suffix: str
            The tplot variable names will be given this suffix.
            Default: None

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.
            Default: None (all variables loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded.  Default: [] (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache.
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword.
            Default: False

        ror: bool
            If set, print PI info and rules of the road.
            Default: True

        uname: str
            User name.  Default: None

        passwd: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> mepi_tof_vars = pyspedas.projects.erg.mepi_tof(trange=['2017-03-27', '2017-03-28'])

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    file_res = 3600. * 24
    prefix = 'erg_mepi_'+level+'_tof'+datatype+'_'

    pathformat = 'satellite/erg/mepi/'+level+'/tof/%Y/%m/erg_mepi_' + \
        level+'_tof'+datatype+'_%Y%m%d_v??_??.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)

    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)

            # --- print PI info and rules of the road

            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: ",gatt["PI_AFFILIATION"])
            print('')
            print('- The rules of the road (RoR) common to the ERG project:')
            print(
                '      https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                '- RoR for MEP-i data: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Mepi')
            print('')
            print('Contact: erg_mep_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if 'flux' in datatype:
        original_suffix_list = ['FPDU', 'FHE2DU', 'FHEDU', 'FOPPDU', 'FODU', 'FO2PDU',
                                'count_raw_P', 'count_raw_HE2', 'count_raw_HE', 'count_raw_OPP', 'count_raw_O', 'count_raw_O2P']
        tplot_names_list = []
        for i in range(len(original_suffix_list)):
            tplot_names_list.append(prefix + original_suffix_list[i] + suffix)
            if tplot_names_list[i] in loaded_data:
                ylim(tplot_names_list[i], 4, 190)

        # set spectrogram plot option
        options(tplot_names_list, 'Spec', 1)

        # set y axis to logscale
        options(tplot_names_list, 'ylog', 1)

        # set ysubtitle
        options(tplot_names_list, 'ysubtitle', '[keV/q]')

        # set ztitle
        options(tplot_names_list[:6], 'ztitle', '[/s-cm^{2}-sr-keV/q]')
        options(tplot_names_list[6:], 'ztitle', '[cnt/smpl]')

        # set z axis to logscale
        options(tplot_names_list, 'zlog', 1)

    elif 'raw' in datatype:

        # set spectrogram plot option
        options(loaded_data, 'Spec', 1)

        # set z axis to logscale
        options(loaded_data, 'zlog', 1)

    return loaded_data



from functools import partial
from .load import load
from pyspedas.utilities.datasets import find_datasets

# Define another name for load
data = load

# retrieve datasets from the ACE mission with a focus on OMNI instrument data
datasets = partial(find_datasets, mission="ACE", instrument="OMNI")



"""
Get information and download files from CDAWeb using cdasws.

For cdasws documentation, see:
    https://pypi.org/project/cdasws/
    https://cdaweb.gsfc.nasa.gov/WebServices/REST/py/cdasws/index.html

"""

import logging
import os
import re
from cdasws import CdasWs
from pyspedas.tplot_tools import cdf_to_tplot, netcdf_to_tplot, time_clip as tclip
from pyspedas.utilities.download import download
from pyspedas.cdagui_tools.config import CONFIG


class CDAWeb:
    """Get information and download files from CDAWeb using cdasws."""

    def __init__(self):
        """Initialize."""
        self.cdas = CdasWs(endpoint=CONFIG["cdas_endpoint"])

    def get_observatories(self):
        """Return a list of strings CDAWeb uses to designate missions or mission groups

        Examples
        --------

        >>> from pyspedas import CDAWeb
        >>> cdaweb_obj = CDAWeb()
        >>> obs_names = cdaweb_obj.get_observatories()
        """
        observatories = self.cdas.get_observatory_groups()
        onames = []
        for mission in observatories:
            mission_name = mission["Name"].strip()
            if len(mission_name) > 1 and mission_name != "(null)":
                onames.append(mission_name)
        return onames

    def get_instruments(self):
        """Return a list of strings CDAWeb uses to designate instrument or dataset types.

        Examples
        --------

        >>> from pyspedas import CDAWeb
        >>> cdaweb_obj = CDAWeb()
        >>> obs_names = cdaweb_obj.get_instruments()
        """
        instruments = self.cdas.get_instrument_types()
        inames = []
        for instrument in instruments:
            instr_name = instrument["Name"].strip()
            if len(instr_name) > 1 and instr_name != "(null)":
                inames.append(instr_name)
        return inames

    def clean_time_str(self, t):
        """Remove the time part from datetime variable."""
        t0 = re.sub("T.+Z", "", t)
        return t0

    def get_datasets(self, mission_list, instrument_list):
        """Return a list of datasets recognized by CDAWeb, given lists of missions and instruments.

        Parameters
        ----------
        mission_list: list of str
            List of mission names, as obtained from get_observatories()
        instrument_list: list of str
            List of instrument names, as obtained from get_instruments()

        Returns
        -------
        list of str
            A list of available datasets for the given missions and instruments.

        Examples
        --------

        >>> from pyspedas import CDAWeb
        >>> cdaweb_obj = CDAWeb()
        >>> dataset_list = cdaweb_obj.get_datasets(['ARTEMIS'],['Electric Fields (space)'])

        """
        thisdict = {"observatoryGroup": mission_list, "instrumentType": instrument_list}
        datasets = self.cdas.get_datasets(**thisdict)
        dnames = []
        for dataset in datasets:
            data_item = dataset["Id"].strip()
            if len(data_item) > 0 and data_item != "(null)":
                tinterval = dataset["TimeInterval"]
                t1 = tinterval["Start"].strip()
                t2 = tinterval["End"].strip()
                t1 = self.clean_time_str(t1)
                t2 = self.clean_time_str(t2)
                data_item += " (" + t1 + " to " + t2 + ")"
            dnames.append(data_item)
        return dnames

    def get_filenames(self, dataset_list, t0, t1):
        """Return a list of urls for a dataset between dates t0 and t1.

        Example: get_files(['THB_L2_FIT (2007-02-26 to 2020-01-17)'],
            '2010-01-01 00:00:00', '2010-01-10 00:00:00')

        Parameters
        ----------
        dataset_list: list of str
            A list of dataset names, as obtained from get_datasets()
        t0: str
            Start time for data to be retrieved
        t1: str
            End time for data to be retrieved

        Returns
        -------
        list of str
            A list of URLs for the given dataset and time range

        Examples
        --------

        >>> from pyspedas import CDAWeb
        >>> cdaweb_obj = CDAWeb()
        >>> urllist = cdaweb_obj.get_filenames(['THB_L2_FIT (2007-02-26 to 2020-01-17)'], '2010-01-01 00:00:00', '2010-01-10 00:00:00')
        """
        remote_url = []

        # Set times to cdas format
        t0 = t0.strip().replace(" ", "T", 1)
        if len(t0) == 10:
            t0 += "T00:00:01Z"
        elif len(t0) > 10:
            t0 += "Z"
        t1 = t1.strip().replace(" ", "T", 1)
        if len(t1) == 10:
            t1 += "T23:23:59Z"
        elif len(t1) > 10:
            t1 += "Z"

        # For each dataset, find the url of files
        for d in dataset_list:
            d0 = d.split(" ")
            if len(d0) > 0:
                status, result = self.cdas.get_data_file(d0[0], [], t0, t1)
                if status == 200 and (result is not None):
                    r = result.get("FileDescription")
                    if r is not None:
                        for f in r:
                            remote_url.append(f.get("Name"))
        return remote_url

    def cda_download(
        self,
        remote_files,
        local_dir=None,
        download_only=False,
        varformat=None,
        get_support_data=False,
        prefix="",
        suffix="",
        varnames=None,
        notplot=False,
        merge=False,
        trange=None,
        time_clip=False,
        force_download=False,
    ):
        """Download data files and (by default) load the data into tplot variables

        Parameters
        ----------
        remote_files : list of str
            List of remote file URLs, as obtained from function get_datasets().
        local_dir : str
            Local directory to save the data in.
        download_only : bool
            If True, download the data, but do not load it into tplot variables.
        varformat: str
            If set, specifies a pattern for which CDF or NetCDF variables to load.
        get_support_data: bool
            If True, load CDF variables marked as 'support_data'.
        prefix: str
            If set, prepend this string to the variable name when creating the tplot variables.
        suffix: str
            If set, append this string to the variable name when creating the tplot variables.
        varnames: list of str
            If set, specifies a list of variables to load from the data files.
            If None or [] or ['*'], load all variables.
        notplot: bool
            If True, return data directly as tplot data structures, rather than a list of tplot names.
        merge: bool
            If True, merge the data with existing tplot variables.
            If False (the default), overwrite existing tplot variables.
        trange: list of str
            If set, clip the time range of the data to these values.
        time_clip: bool
            If True, clip the time range of the data to the values in trange.
        force_download: bool
            If True, download the data even if it already exists locally.

        Returns
        -------
        tuple
            A tuple containing the number of files downloaded, the number of variables loaded, and a list of the tplot variables loaded.

        Examples
        --------
        >>> from pyspedas import CDAWeb
        >>> from pyspedas import tplot
        >>> cdaweb_obj = CDAWeb()
        >>> urllist = cdaweb_obj.get_filenames(['THB_L2_FIT (2007-02-26 to 2020-01-17)'], '2010-01-01 00:00:00', '2010-01-10 00:00:00')
        >>> result = cdaweb_obj.cda_download(urllist,local_dir="/tmp")
        >>> tplot('thb_fgs_gsm')
        """

        # Return quantities
        no_of_files = 0
        no_of_variables = 0
        loaded_vars = []

        # Set the local and remote directories
        remotehttp = CONFIG["remote_data_dir"]
        if local_dir is None:
            local_dir = CONFIG["local_data_dir"]

        cdf_files = []
        netcdf_files = []
        all_files = []

        # Download the files
        for remotef in remote_files:
            f = remotef.strip().replace(remotehttp, "", 1)
            localf = os.path.normpath(local_dir + os.path.sep + f)
            localfiles = download(
                remote_file=remotef,
                local_file=localf,
                force_download=force_download,
            )
            if localfiles is None:
                continue
            for f in localfiles:
                if f is not None and len(f) > 0:
                    all_files.append(os.path.normpath(f))

        no_of_files = len(all_files)
        if no_of_files > 0:

            # Sort the file list
            all_files = list(set(all_files))
            all_files.sort()

            # Load the data into tplot variables
            if not download_only:
                # Separate cdf and netcdf files. All other files cannot be loaded into tplot.
                for f in all_files:
                    if f.endswith(".cdf"):
                        cdf_files.append(f)
                    elif f.endswith(".nc"):
                        netcdf_files.append(f)
                    else:
                        logging.warning("File type not supported: %s", f)

                if len(cdf_files) > 0:
                    cdf_files.sort()
                    logging.info("Downloaded %d CDF files.", len(cdf_files))
                    try:
                        cdf_vars = cdf_to_tplot(
                            cdf_files,
                            prefix=prefix,
                            suffix=suffix,
                            get_support_data=get_support_data,
                            varformat=varformat,
                            varnames=varnames,
                            notplot=notplot,
                            merge=merge,
                        )
                        if cdf_vars is not None:
                            loaded_vars.extend(cdf_vars)
                    except ValueError as err:
                        msg = "cdf_to_tplot could not load " + str(cdf_files)
                        msg += "\n\n"
                        msg += "Error from pytplot: " + str(err)
                        logging.error(msg)

                if len(netcdf_files) > 0:
                    netcdf_files.sort()
                    logging.info("Downloaded %d NetCDF files.", len(netcdf_files))
                    try:
                        netcdf_vars = netcdf_to_tplot(
                            netcdf_files,
                            prefix=prefix,
                            suffix=suffix,
                            merge=merge,
                        )
                        if netcdf_vars is not None:
                            loaded_vars.extend(netcdf_vars)
                    except ValueError as err:
                        msg = "netcdf_to_tplot could not load " + str(netcdf_files)
                        msg += "\n\n"
                        msg += "Error from pytplot: " + str(err)
                        logging.error(msg)

                loaded_vars = list(set(loaded_vars))
                no_of_variables = len(loaded_vars)
                logging.info("Number of tplot variables loaded: %d", no_of_variables)

                if time_clip and trange is not None:
                    if trange[0] >= trange[1]:
                        logging.warning(
                            "trange values equal or out of order, no time clipping performed"
                        )
                    else:
                        tclip(
                            loaded_vars, trange[0], trange[1], suffix="", overwrite=True
                        )
                elif time_clip:
                    logging.warning("Warning: No trange specified for time_clip")

        return (no_of_files, no_of_variables, loaded_vars)



import logging
from .load_dst import dst as load_dst
from .load_ae import load_ae
from pyspedas.projects.themis.ground.gmag import gmag as thm_gmag
from pyspedas.projects.noaa.noaa_load_kp import noaa_load_kp
from pyspedas.projects.omni import data as omni_load
from pyspedas.tplot_tools import time_clip as tclip


def load_geomagnetic_indices(
    datatypes=None,
    omni_load_all=False,
    missions=["kyoto", "themis", "noaa", "gfz", "omni"],
    trange=["2024-05-11 00:00:00", "2024-05-11 23:59:59"],
    prefix="",
    suffix="",
    force_download=False,
    time_clip=True,
):
    """
    Loads geomagnetic indices data from different sources.

    Kyoto: ["dst", "ae", "al", "ao", "au", "ax"]
    Themis: ['thg_idx_al', 'thg_idx_au', 'thg_idx_ae', 'thg_idx_uc_al', 'thg_idx_uc_au', 'thg_idx_uc_ae', 'thg_idx_uc_avg']
    NOAA or gfz: ['Kp', 'ap', 'Sol_Rot_Num', 'Sol_Rot_Day', 'Kp_Sum', 'ap_Mean', 'Cp', 'C9', 'Sunspot_Number', 'F10.7', 'Flux_Qualifier']
    Omni: ['AE_INDEX', 'AL_INDEX', 'AU_INDEX', 'SYM_D', 'SYM_H', 'ASY_D', 'ASY_H', 'Pressure']

    Parameters
    ----------
    datatypes : list, optional
        A list of strings specifying the types of geomagnetic indices to load.
        Default is None, which downloads all available indices from the specified missions.
    missions : list, optional
        A list of strings specifying the data missions to load the indices from.
        Default is ["kyoto", "themis", "noaa", "omni"].
    omni_load_all : bool, optional
        A boolean indicating whether to load all available variables from the Omni data.
        Default is False.
    trange : list or tuple, optional
        A list specifying the time range of the data to load.
        Default is ["2024-05-11 00:00:00", "2024-05-11 23:59:59"].
    prefix : str, optional
        A string to prepend to the variable names when loading the data.
        Default is an empty string.
    suffix : str, optional
        A string to append to the variable names when loading the data.
        Default is an empty string.
    force_download : bool, optional
        A boolean specifying whether to force the data to be downloaded,
        even if the data already exists locally.
        Default is False.
    time_clip : bool, optional
        A boolean specifying whether to clip the data to the specified time range.
        Default is True.

    Returns
    -------
    List of str
        A list of tplot variable names that were loaded.

    """
    vars = []  # list of tplot variables to return

    if datatypes is None or datatypes == [] or "*" in datatypes:
        datatypes = None

    if trange is None or len(trange) != 2:
        logging.error("Keyword trange with two datetimes is required to download data.")
        return vars
    if trange[0] >= trange[1]:
        logging.error("Invalid time range. End time must be greater than start time.")
        return vars

    missions = [missions] if isinstance(missions, str) else missions

    for p in missions:
        p = p.lower()
        if p == "kyoto":
            if datatypes is None or "dst" in datatypes:
                dst_vars = load_dst(
                    trange=trange,
                    prefix=prefix,
                    suffix=suffix,
                    time_clip=time_clip,
                    force_download=force_download,
                )
                if len(dst_vars) < 1:
                    logging.info("No Kyoto Dst data found.")
                else:
                    vars.extend(dst_vars)
            if datatypes is not None and "dst" in datatypes:
                datatypes.remove("dst")

            if datatypes is None:
                ae_vars = load_ae(
                    trange=trange,
                    prefix=prefix,
                    suffix=suffix,
                    time_clip=time_clip,
                    force_download=force_download,
                )
                if len(ae_vars) < 1:
                    logging.info("No Kyoto AE data found.")
                else:
                    vars.extend(ae_vars)
            else:
                ae_vars = load_ae(
                    trange=trange,
                    datatypes=datatypes,
                    prefix=prefix,
                    suffix=suffix,
                    time_clip=time_clip,
                    force_download=force_download,
                )
                if len(ae_vars) < 1:
                    logging.info("No Kyoto AE data found.")
                else:
                    vars.extend(ae_vars)
        elif p == "themis":
            # Todo: currently, gmag does not support prefix
            if datatypes is None:
                thm_vars = thm_gmag(
                    trange=trange,
                    sites="idx",
                    prefix=prefix,
                    suffix=suffix,
                    time_clip=time_clip,
                    force_download=force_download,
                )
                if len(thm_vars) < 1:
                    logging.info("No THEMIS index data found.")
                else:
                    vars.extend(thm_vars)
            else:
                thm_vars = thm_gmag(
                    trange=trange,
                    sites="idx",
                    varnames=datatypes,
                    prefix=prefix,
                    suffix=suffix,
                    time_clip=time_clip,
                    force_download=force_download,
                )
                if len(thm_vars) < 1:
                    logging.info("No THEMIS index data found.")
                else:
                    vars.extend(thm_vars)
        elif p == "noaa":
            nprefix = prefix + "noaa_"  # NOAA data needs a prefix
            if datatypes is None:
                noaa_vars = noaa_load_kp(
                    trange=trange,
                    prefix=nprefix,
                    suffix=suffix,
                    time_clip=time_clip,
                    force_download=force_download,
                )
                if len(noaa_vars) < 1:
                    logging.info("No NOAA index data found.")
                else:
                    vars.extend(noaa_vars)
            else:
                noaa_vars = noaa_load_kp(
                    trange=trange,
                    datatype=datatypes,
                    prefix=nprefix,
                    suffix=suffix,
                    time_clip=time_clip,
                    force_download=force_download,
                    gfz=False,
                )
                if len(noaa_vars) < 1:
                    logging.info("No NOAA index data found.")
                else:
                    vars.extend(noaa_vars)
        elif p == "gfz":
            gprefix = prefix + "gfz_"  # GFZ data needs a prefix
            if datatypes is None:
                gfz_vars = noaa_load_kp(
                    trange=trange,
                    prefix=gprefix,
                    suffix=suffix,
                    time_clip=time_clip,
                    gfz=True,
                    force_download=force_download,
                )
                if len(gfz_vars) < 1:
                    logging.info("No GFZ index data found.")
                else:
                    vars.extend(gfz_vars)
            else:
                gfz_vars = noaa_load_kp(
                    trange=trange,
                    datatype=datatypes,
                    prefix=gprefix,
                    suffix=suffix,
                    time_clip=time_clip,
                    gfz=True,
                )
                if len(gfz_vars) < 1:
                    logging.info("No GFZ index data found.")
                else:
                    vars.extend(gfz_vars)
        elif p == "omni":
            oprefix = prefix + "omni_"  # Omni data needs a prefix
            if datatypes is None and not omni_load_all:
                # Omni loads too many variables, we need to restrict them
                odatatypes = [
                    "Pressure",
                    "AE_INDEX",
                    "AL_INDEX",
                    "AU_INDEX",
                    "SYM_D",
                    "SYM_H",
                    "ASY_D",
                    "ASY_H",
                ]
            else:
                odatatypes = ['IMF', 'PLS', 'IMF_PTS', 'PLS_PTS', 'percent_interp', 'Timeshift',
                    'RMS_Timeshift', 'RMS_phase', 'Time_btwn_obs', 'F', 'BX_GSE', 'BY_GSE', 'BZ_GSE',
                    'BY_GSM', 'BZ_GSM', 'RMS_SD_B', 'RMS_SD_fld_vec', 'flow_speed', 'Vx',  'Vy', 'Vz',
                    'proton_density', 'T','NaNp_Ratio', 'Pressure', 'E', 'Beta', 'Mach_num', 'Mgs_mach_num',
                    'x', 'y', 'z', 'BSN_x', 'BSN_y', 'BSN_z', 'AE_INDEX', 'AL_INDEX', 'AU_INDEX', 'SYM_D',
                     'SYM_H', 'ASY_D','ASY_H']
            omni_vars = omni_load(
                trange=trange,
                prefix=oprefix,
                suffix=suffix,
                time_clip=time_clip,
                varnames=odatatypes,
                force_download=force_download,
            )
            if len(omni_vars) < 1:
                logging.info("No Omni data found.")
            else:
                vars.extend(omni_vars)

    if time_clip:
        tclip(vars, trange[0], trange[1], overwrite=True)

    return vars



import os

CONFIG = {
    "local_data_dir": "goes_data/",
    "remote_data_dir": "https://www.ncei.noaa.gov/data/goes-space-environment-monitor/access/",
}

# override local data directory with environment variables
if os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.sep.join([os.environ["SPEDAS_DATA_DIR"], "goes"])

if os.environ.get("GOES_DATA_DIR"):
    CONFIG["local_data_dir"] = os.environ["GOES_DATA_DIR"]



import logging
import pandas as pd
import numpy as np
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import store_data, options


def mms_load_qf_tplot(filenames, suffix='', trange=None):
    """
    Helper routine for loading tetrahedron quality factor data (ASCII files from the SDC); not meant to be called directly
    """
    prefix = 'mms'

    time_values = []
    date_values = []
    qf_values = []

    for file in filenames:
        logging.info('Loading ' + file)
        rows = pd.read_csv(file, sep=r'\s+', header=None, skiprows=11)
        times = rows.shape[0]
        for time_idx in range(0, times):
            # these files can overlap, so avoid duplicates
            if rows[0][time_idx] in date_values:
                continue
            time_values.append(pd.to_datetime(rows[0][time_idx], format='%Y-%j/%H:%M:%S.%f').timestamp())
            qf_values.append(rows[2][time_idx])
            date_values.append(rows[0][time_idx])

    tp_name=prefix + '_' + 'tetrahedron_qf' + suffix
    store_data(tp_name, data={'x': np.array(time_values), 'y': np.array(qf_values)})
    tclip(tp_name, trange[0], trange[1], suffix='')
    options(tp_name, 'ytitle', 'MMS Tetrahedron Quality Factor')
    options(tp_name, 'legend_names', 'qf')
    return [tp_name]




import cdflib
import numpy as np

from copy import deepcopy
from pyspedas.tplot_tools import get_data, store_data, options, clip, ylim

from ...satellite.erg.load import load
from ...satellite.erg.get_gatt_ror import get_gatt_ror


from typing import List, Optional, Union


def gmag_isee_induction(
    trange: List[str] = ["2018-10-18/00:00:00", "2018-10-18/02:00:00"],
    suffix: str = "",
    site: Union[str, List[str]] = "all",
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    frequency_dependent: bool = False,
    force_download: bool = False,
) -> List[str]:
    """
    Load data from ISEE Induction Magnetometers

    Parameters
    ----------
    trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2018-10-18/00:00:00', '2018-10-18/02:00:00']

    suffix: str
            The tplot variable names will be given this suffix.  Default: ''

    site: str or list of str
            The site or list of sites to load. Valid values: "ath", "gak", "hus", "kap", "lcl", "mgd", "msr",
            "nai", "ptk", "rik", "sta", "zgn", "all"
            Default: ['all']

    get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: False

    varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

    varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

    downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

    notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

    no_update: bool
            If set, only load data from your local cache. Default: False

    uname: str
            User name.  Default: None

    passwd: str
            Password. Default: None

    time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

    ror: bool
            If set, print PI info and rules of the road. Default: True

    frequency_dependent: bool
            If set, load frequency dependent data.
            Default: False

    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------

    Examples
    ________

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> ind_vars=pyspedas.projects.erg.gmag_isee_induction(trange=['2020-08-01','2020-08-02'], site='all')
    >>> tplot('isee_induction_db_dt_msr')
    """

    site_code_all = [
        "ath",
        "gak",
        "hus",
        "kap",
        "lcl",
        "mgd",
        "msr",
        "nai",
        "ptk",
        "rik",
        "sta",
        "zgn",
    ]

    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(" ")
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if "all" in site_code:
        site_code = site_code_all

    site_code = list(set(site_code).intersection(site_code_all))

    if frequency_dependent:
        frequency_dependent_structure = {}
        for site_input in site_code:
            frequency_dependent_structure[site_input] = {
                "site_code": "",
                "nfreq": 0,
                "frequency": np.zeros(shape=(64)),
                "sensitivity": np.zeros(shape=(64, 3)),
                "phase_difference": np.zeros(shape=(64, 3)),
            }

    prefix = "isee_induction_"
    if notplot:
        loaded_data = {}
    else:
        loaded_data = []

    for site_input in site_code:
        file_res = 3600.0
        pathformat = (
            "ground/geomag/isee/induction/"
            + site_input
            + "/%Y/%m/isee_induction_"
            + site_input
            + "_%Y%m%d%H_v??.cdf"
        )

        loaded_data_temp = load(
            pathformat=pathformat,
            file_res=file_res,
            trange=trange,
            prefix=prefix,
            suffix="_" + site_input + suffix,
            get_support_data=get_support_data,
            varformat=varformat,
            downloadonly=downloadonly,
            notplot=notplot,
            time_clip=time_clip,
            no_update=no_update,
            uname=uname,
            passwd=passwd,
            force_download=force_download
        )

        if notplot:
            loaded_data.update(loaded_data_temp)
        else:
            loaded_data += loaded_data_temp
        if (len(loaded_data_temp) > 0) and ror:
            try:
                gatt = get_gatt_ror(downloadonly, loaded_data)
                print(
                    "**************************************************************************"
                )
                print(gatt["Logical_source_description"])
                print("")
                print(f'Information about {gatt["Station_code"]}')
                print("PI and Host PI(s):")
                print(gatt["PI_name"])
                print("")
                print("Affiliation: ")
                print(gatt["PI_affiliation"])
                print("")
                print("Rules of the Road for ISEE Induction Magnetometer Data Use:")
                for gatt_text in gatt["TEXT"]:
                    print(gatt_text)
                print(gatt["LINK_TEXT"])
                print(
                    "**************************************************************************"
                )
            except:
                print("printing PI info and rules of the road was failed")

        if (not downloadonly) and (not notplot):
            tplot_name = prefix + "db_dt_" + site_input + suffix
            if tplot_name in loaded_data:
                clip(tplot_name, -1e4, 1e4)
                get_data_vars = get_data(tplot_name)
                ylim(
                    tplot_name, np.nanmin(get_data_vars[1]), np.nanmax(get_data_vars[1])
                )
                options(tplot_name, "legend_names", ["dH/dt", "dD/dt", "dZ/dt"])
                options(tplot_name, "Color", ["b", "g", "r"])
                options(tplot_name, "ytitle", "\n".join(tplot_name.split("_")))

                if frequency_dependent:
                    meta_data_var = get_data(tplot_name, metadata=True)
                    new_cdflib = False
                    if cdflib.__version__ > "0.4.9":
                        new_cdflib = True
                    else:
                        new_cdflib = False

                    if meta_data_var is not None:
                        cdf_file = cdflib.CDF(meta_data_var["CDF"]["FILENAME"])
                        if new_cdflib:
                            cdfcont = cdf_file.varinq("frequency")
                        else:
                            cdfcont = cdf_file.varget("frequency", inq=True)
                        ffreq = cdf_file.varget("frequency")
                        ssensi = cdf_file.varget("sensitivity")
                        pphase = cdf_file.varget("phase_difference")
                        frequency_dependent_structure[site_input][
                            "site_code"
                        ] = site_input
                        if new_cdflib:
                            frequency_dependent_structure[site_input]["nfreq"] = cdfcont.Last_Rec + 1
                        else:
                            frequency_dependent_structure[site_input]["nfreq"] = cdfcont["max_records"] + 1
                        frequency_dependent_structure[site_input]["frequency"][
                            0 : ffreq.shape[0]
                        ] = deepcopy(ffreq)
                        frequency_dependent_structure[site_input]["sensitivity"][
                            0 : ssensi.shape[0]
                        ] = deepcopy(ssensi)
                        frequency_dependent_structure[site_input]["phase_difference"][
                            0 : pphase.shape[0]
                        ] = deepcopy(pphase)

    if frequency_dependent:
        return frequency_dependent_structure
    else:
        return loaded_data



from pyspedas import options, ylim, zlim, get_data

from ..load import load
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional

def pwe_ofa(
    trange: List[str] = ['2017-04-01', '2017-04-02'],
    datatype: str = 'spec',
    level: str = 'l2',
    suffix: str = '',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the PWE experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-04-01', '2017-04-02']

        datatype: str
            Data type; Valid 'l2' options: 'complex', 'matrix', 'spec'. Valid 'l3' options: 'property'
            Default: 'spec'

        level: str
            Data level; Valid options: 'l2', 'l3'
            Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix. Default: None

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded. Default: [] (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: True

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        ror: bool
            If set, print PI info and rules of the road
            Default: True

        uname: str
            User name. Default: None

        passwd: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> pwe_ofa_vars = pyspedas.projects.erg.pwe_ofa(trange=['2017-03-27', '2017-03-28'])
    >>> tplot('erg_pwe_ofa_l2_spec_E_spectra_132')

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    file_res = 3600. * 24
    prefix = 'erg_pwe_ofa_'+level+'_'+datatype+'_'

    pathformat = 'satellite/erg/pwe/ofa/'+level+'/'+datatype + \
        '/%Y/%m/erg_pwe_ofa_'+level+'_'+datatype+'_%Y%m%d_v??_??.cdf'

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)

    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)

            # --- print PI info and rules of the road

            print(' ')
            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG PWE OFA')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: ", gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                'RoR of PWE/OFA: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Pwe/Ofa')
            print('')
            print('Contact: erg_pwe_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    # set spectrogram plot option
    options(prefix+'E_spectra_132'+suffix,  'Spec', 1)
    options(prefix+'B_spectra_132'+suffix,  'Spec', 1)

    # set y axis to logscale
    options(prefix+'E_spectra_132'+suffix,  'ylog', 1)
    options(prefix+'B_spectra_132'+suffix,  'ylog', 1)

    if prefix+'E_spectra_132'+suffix in loaded_data:
        # set ylim
        ylim(prefix+'E_spectra_132'+suffix, 32e-3, 20.)
        # set zlim
        zlim(prefix+'E_spectra_132'+suffix, 1e-9, 1e-2)

    if prefix+'B_spectra_132'+suffix in loaded_data:
        # set ylim
        ylim(prefix+'B_spectra_132'+suffix, 32e-3, 20.)
        # set zlim
        zlim(prefix+'B_spectra_132'+suffix, 1e-4, 1e2)

    # set ytitle
    options(prefix+'E_spectra_132'+suffix,  'ytitle', 'ERG PWE/OFA-SPEC (E)')
    options(prefix+'B_spectra_132'+suffix,  'ytitle', 'ERG PWE/OFA-SPEC (B)')

    # set ysubtitle
    options(prefix+'E_spectra_132'+suffix,  'ysubtitle', 'frequency [kHz]')
    options(prefix+'B_spectra_132'+suffix,  'ysubtitle', 'frequency [kHz]')

    # set ztitle
    options(prefix+'E_spectra_132'+suffix,  'ztitle', 'mV^2/m^2/Hz')
    options(prefix+'B_spectra_132'+suffix,  'ztitle', 'pT^2/Hz')

    # set z axis to logscale
    options(prefix+'E_spectra_132'+suffix,  'zlog', 1)
    options(prefix+'B_spectra_132'+suffix,  'zlog', 1)

    # change colormap option
    options(prefix+'E_spectra_132'+suffix,  'Colormap', 'jet')
    options(prefix+'B_spectra_132'+suffix,  'Colormap', 'jet')

    return loaded_data



import urllib.request
import csv
import pyspedas
from io import StringIO
from collections import defaultdict
from pyspedas.tplot_tools import time_datetime


def _list_of_fdsn_channels():
    """
        Generates a list of channel names based on predefined bands, orientations, and a single instrument type.

        Returns
        -------
        str
            A string containing channel names, each separated by a comma. T

        Examples
        --------
        >>> _list_of_fdsn_channels()
        'MFZ,MFN,MFE,LFZ,LFN,LFE,VFZ,VFN,VFE,UFZ,UFN,UFE'
     """

    channel_band = ['M', 'L', 'V', 'U']
    channel_orientation = ['Z', 'N', 'E']
    # channel_orientation = ['Z', 'N', 'E', '1', '2'] # Some of the data can be stored as '1', '2', which is not supported byt this tool.
    channel_instrument = 'F'
    # Create list of all combinations of channel_band, channel_instrument and channel_orientation
    channel_list = ",".join(
        [f"{band}{channel_instrument}{orient}" for band in channel_band for orient in channel_orientation])
    return channel_list


def mth5_time_str(time):
    """
        Return pyspedas time format as  YYYY-MM-DDThh:mm:ss

        Parameters:
            time :  str
                Time in pyspedas format

        Returns:
            Time in MTH5 format

    """
    return time_datetime(time).strftime("%Y-%m-%dT%H:%M:%S")


def datasets(trange=None, network=None, station=None, USAarea=False):
    """
    Fetches datasets availablity based on time range, network, and station.

    The function queries an FDSN (Federation of Digital Seismograph Networks) web service for station
    data within a specified time range and geographical bounds limited to the USA. It filters the resulting channels
    based on selection of 3 chanles of the same band and diffrent orienation, e.g., 'LFE', 'LFN', 'LFZ'. If the dataset has
    less than 3 channels with the same band, it is excluded. The resulting dataset is organized by network, station, and time
    range, including only the channels that meet the specified criteria.

    Parameters
    ----------
    trange : list of str
         Time range of interest [starttime, endtime] with the format ['YYYY-MM-DD','YYYY-MM-DD']
    network : str, optional
        The network code to filter the datasets. If None (default), no network filter is applied.
    station : str, optional
        The station code to filter the datasets. If None (default), no station filter is applied.
    USAarea : bool, optional
        If True, restricts the search to the geographical boundaries (box) of the USA:

        - Latitude: 24 to 49 degrees
        - Longitude: -127 to -59 degrees

    Returns
    -------
    dict
        A dictionary organized by network and station, each containing a dictionary of time ranges
        (start_date, end_date) mapping to a list of channels that meet the filter criteria. The
        dictionary is empty if no channels meet the criteria or if the inputs do not match any
        available data.

    Examples
    --------
    >>> datasets(["2015-06-22", "2015-06-23"], network="4P", station="ALW48")
    {'4P': {'ALW48': {('2015-06-18T15:00:36.0000', '2015-07-09T13:45:10.0000'): ['LFE', 'LFN', 'LFZ']}}}

    Note
    ----
    The function requires internet access to query the FDSN web service and parse the returned data.
    It uses the `urllib` library for the web request and the `csv` module to parse the response.
    """

    if not trange:
        return None

    # Function to filter channels based on the starting letter criteria
    def filter_channels_by_letter(channels):
        # Count channels starting with each letter
        channel_start_count = defaultdict(int)
        for channel in channels:
            if channel:  # Ensure the channel is not empty
                channel_start_count[channel[0]] += 1

        # Filter channels where the starting letter count is 3 or more
        return any(count >= 3 for count in channel_start_count.values())

    # Get data and handle empty response
    def fetch_data(url):
        try:
            with urllib.request.urlopen(url) as response:
                if response.status == 200:
                    return response.read().decode('utf-8')  # Or any other processing
                else:
                    pyspedas.logger.error("Server returned status code: %d", response.status)
        except (urllib.error.HTTPError, urllib.error.URLError) as e:
            if hasattr(e, 'code') and e.code == 404:
                # No data
                return None
            else:
                pyspedas.logger.error("HTTP or URL Error encountered: %s", e)
        return None  # or an appropriate fallback value/error indicator

    t1 = mth5_time_str(trange[0])  # ["2019-11-14T00:00:00"],
    t2 = mth5_time_str(trange[1])  # ["2019-11-15T00:00:00"]

    # FDSN F Channels
    cha = _list_of_fdsn_channels()

    net_str = ''
    sta_str = ''
    if network:
        net_str = f'net={network}'

    if station:
        sta_str = f'sta={station}'

    if USAarea:
        # URL of the data limited to USA coordinate box
        # lat 24 49
        # lon -127 -59
        url = f'https://service.iris.edu/fdsnws/station/1/query?{net_str}&{sta_str}&loc=*&cha={cha}&starttime={t1}&endtime={t2}&level=channel&format=text&maxlat=49&minlon=-127&maxlon=-59&minlat=24&includecomments=false&nodata=404'
    else:
        # URL of the data unlimited in location
        url = f'https://service.iris.edu/fdsnws/station/1/query?{net_str}&{sta_str}&loc=*&cha={cha}&starttime={t1}&endtime={t2}&level=channel&format=text&includecomments=false&nodata=404'

    # Initialize the result dictionary
    res = {}

    # Fetching the data
    response_text = fetch_data(url)
    if response_text is not None:
        response_io = StringIO(response_text)

        reader = csv.reader(response_io, delimiter='|', skipinitialspace=True)
        reader.__next__()  # Skip the header row

        # Iterate over the rows in the reader object
        for row in reader:
            network, station, _, channel, *_, start_date, end_date = row

            # Initialize the network dictionary if not present
            if network not in res:
                res[network] = {}

            # Initialize the station dictionary within the network if not present
            if station not in res[network]:
                res[network][station] = {}

            # Use (start_date, end_date) tuple as a key, mapping to a list of channels
            date_tuple = (start_date, end_date)
            if date_tuple not in res[network][station]:
                res[network][station][date_tuple] = []

            # Append the current channel to the list for this date_tuple
            res[network][station][date_tuple].append(channel)
    else:
        return res

    # Filter the constructed dictionary, using copies of the iterators
    for network in list(res.keys()):
        for station in list(res[network].keys()):
            for date_tuple in list(res[network][station].keys()):
                channels = res[network][station][date_tuple]
                if not filter_channels_by_letter(channels):
                    del res[network][station][date_tuple]

            # Remove the station if it becomes empty after the filtering
            if not res[network][station]:
                del res[network][station]

        # Remove the network if it becomes empty after the filtering
        if not res[network]:
            del res[network]

    return res



import gzip


def is_gzip(file_path):
    r"""
    Check if a file is gzip-compressed.

    Parameters
    ----------
    file_path : str
       Path to the file to be checked

    Returns
    -------
    bool
        True if first two bytes are '0x1f 0x8b', false otherwise

    """
    with open(file_path, 'rb') as f:
        return f.read(2) == b'\x1f\x8b'
